import glob
import os
import re
from pathlib import Path
from typing import Callable
from unittest.mock import MagicMock, patch
from urllib.parse import urljoin

import click

import pytest

import toml
from freezegun import freeze_time

from helpers import assert_approx_duration

from ytpb.config import DEFAULT_CONFIG


@pytest.mark.parametrize(
    "interval,output_subpath",
    [
        # Segments and corresponding ingestion start dates:
        #             7959120       21       22
        #                  |        |        |
        # 2023-03-25T23:33:54.491Z  56.490   58.492
        ("7959120/7959121", "233354+00.mp4"),
        ("7959120/2023-03-25T23:33:57+00", "233354+00.mp4"),
        ("7959120/PT3S", "233354+00.mp4"),
        ("2023-03-25T23:33:55+00/2023-03-25T23:33:57+00", "233355+00.mp4"),
        ("2023-03-25T23:33:55+00/PT3S", "233355+00.mp4"),
        ("2023-03-25T23:33:55+00/7959121", "233355+00.mp4"),
        ("PT3S/7959121", "233355+00.mp4"),
        ("PT3S/2023-03-25T23:33:58+00", "233355+00.mp4"),
    ],
)
@freeze_time("2023-03-26T00:00:00+00:00")
def test_download_within_interval(
    interval: str,
    output_subpath: str,
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    expected_out,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                interval,
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
            catch_exceptions=False,
            standalone_mode=False,
        )

    # Then:
    assert result.output == expected_out
    assert result.exit_code == 0
    assert glob.glob(str(tmp_path / f"*{output_subpath}"))


@freeze_time("2023-03-26T00:00:00+00:00")
def test_providing_start_and_end_equals_now(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:

    # In this test the reference segment refers to a segment (7959122) next to
    # the end one (7959121).
    add_responses_callback_for_reference_base_url(7959122)
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/now",  # 'now' should refer to segment 7959121
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    expected_path = tmp_path / "Webcam-Zurich-HB_20230325T233354+00.mp4"
    assert os.path.exists(expected_path)
    assert_approx_duration(expected_path, 4.0)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_preview_option_with_open_interval(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # We only have three test segments of 2 second duration.
    DEFAULT_CONFIG["general"]["preview_duration"] = 4

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "2023-03-25T23:33:55+00/..",
                "--preview",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    expected_path = tmp_path / "Webcam-Zurich-HB_20230325T233355+00.mp4"
    assert os.path.exists(expected_path)
    assert_approx_duration(expected_path, 5.5)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_preview_option_with_closed_interval(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # We only have three test segments of 2 second duration.
    DEFAULT_CONFIG["general"]["preview_duration"] = 4

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "2023-03-25T23:33:55+00/2023-03-25T23:33:57:00+00",
                "--preview",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
            catch_exceptions=False,
        )

    # Then:
    assert result.exit_code == 0

    expected_path = tmp_path / "Webcam-Zurich-HB_20230325T233355+00.mp4"
    assert os.path.exists(expected_path)
    assert_approx_duration(expected_path, 5.5)

    expected = "info: The preview mode is enabled, interval end is ignored."
    assert expected in result.output


@pytest.mark.parametrize(
    "audio_format,video_format",
    [("itag eq 140", "itag eq 244"), ("itag eq 140", "none"), ("none", "itag eq 244")],
)
@freeze_time("2023-03-26T00:00:00+00:00")
def test_download_audio_and_or_video(
    audio_format: str | None,
    video_format: str | None,
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    video_base_url: str,
    run_temp_directory: Path,
    expected_out,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
        urljoin(video_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--no-cut",
                "--interval",
                "7959120/7959121",
                "-af",
                audio_format,
                "-vf",
                video_format,
                "--no-cleanup",
                stream_url,
            ],
            catch_exceptions=False,
        )

    # Then:
    assert result.exit_code == 0
    expected_output = re.sub(
        r"notice: No cleanup enabled, check .+",
        f"notice: No cleanup enabled, check {run_temp_directory}/",
        expected_out._pattern_filename.read_text(),
    )
    assert expected_output == result.output


@freeze_time("2023-03-26T00:00:00+00:00")
def test_download_to_custom_absolute_path(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "-o",
                tmp_path / "test" / "merged",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    assert os.path.exists(tmp_path / "test" / "merged.mp4")


@freeze_time("2023-03-26T00:00:00+00:00")
def test_download_to_custom_relative_path(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "-o",
                "test/merged",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    assert os.path.exists(tmp_path / "test" / "merged.mp4")


@pytest.mark.parametrize(
    "output_path,expected",
    [
        (
            "<title>_<input_start_date>_<duration>",
            "Webcam-Zurich-HB_20230325T233354+00_PT4S.mp4",
        ),
    ],
)
@freeze_time("2023-03-26T00:00:00+00:00")
def test_download_to_template_output_path(
    output_path: str,
    expected: str,
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "-o",
                output_path,
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    assert expected == str(next(tmp_path.glob("*.mp4")).name)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_no_cleanup_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "--no-cleanup",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    assert os.path.exists(tmp_path / "Webcam-Zurich-HB_20230325T233354+00.mp4")
    assert os.path.exists(run_temp_directory / "7959120.i140.mp4")


@freeze_time("2023-03-26T00:00:00+00:00")
def test_no_merge_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
    expected_out,
) -> None:
    """The expected output file for this test contains a template variable, and
    needs to be updated manually."""

    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "--no-merge",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    assert not os.path.exists(tmp_path / "Webcam-Zurich-HB_20230325T233354+00.mp4")
    assert os.path.exists(run_temp_directory / "7959120.i140.mp4")

    expected_output = re.sub(
        r"Success! Segments saved to .+\.",
        f"Success! Segments saved to {run_temp_directory}/segments/.",
        expected_out._pattern_filename.read_text(),
    )
    expected_output = re.sub(
        r"notice: No cleanup enabled, check .+",
        f"notice: No cleanup enabled, check {run_temp_directory}/",
        expected_output,
    )
    assert expected_output == result.output


@freeze_time("2023-03-26T00:00:00+00:00")
def test_dry_run_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
    expected_out,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "--dry-run",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    assert result.output == expected_out
    assert not os.path.exists(tmp_path / "Webcam-Zurich-HB_20230325T233355+00.mp4")
    assert not os.path.exists(run_temp_directory)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_no_cut_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
    expected_out,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "2023-03-25T23:33:55+00/2023-03-25T23:33:58+00",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "--no-cut",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    expected_path = tmp_path / "Webcam-Zurich-HB_20230325T233355+00.mp4"
    assert os.path.exists(expected_path)
    assert_approx_duration(expected_path, 4.0)

    assert result.output == expected_out


@freeze_time("2023-09-28T17:00:00+00:00")
def test_from_cache(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    create_cache_file: None,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
):
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cut",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    assert "Cache for the video doesn't exist" not in result.output


@freeze_time("2023-03-26T00:00:00+00:00")
def test_from_empty_cache(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    run_temp_directory: Path,
):
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cut",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0
    expected_path = tmp_path / "Webcam-Zurich-HB_20230325T233354+00.mp4"
    assert os.path.exists(expected_path)


@pytest.mark.parametrize(
    "audio_format,video_format",
    [
        ("format eq mp4", "quality gt 360p"),
        ("format eq mp4", "itag eq 244"),
        ("itag eq 140", "quality gt 360p"),
        ("format eq mp4", "none"),
        ("none", "quality gt 360p"),
    ],
)
@freeze_time("2023-03-26T00:00:00+00:00")
def test_ambiguous_format_specs(
    audio_format: str,
    video_format: str,
    ytpb_cli_invoke: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
    # expected_out,
) -> None:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cut",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                audio_format,
                "-vf",
                video_format,
                stream_url,
            ],
        )

    assert result.exit_code == 1
    # assert result.output == expected_out


@freeze_time("2023-03-26T00:00:00+00:00")
def test_yt_dlp_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_id: str,
    audio_base_url: str,
    video_base_url: str,
    tmp_path: Path,
) -> None:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    with patch("ytpb.cli.common.YoutubeDLInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--yt-dlp",
                "--no-cut",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
        )

    assert result.exit_code == 0
    expected_path = tmp_path / "Webcam-Zurich-HB_20230325T233354+00.mp4"
    assert os.path.exists(expected_path)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_with_default_config_file(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    """Tests loading of a configuration from the default path. Here two things
    are tested:

    1. The default values for options are loaded from the configuration file.
    2. The command options hold priority over the configuration.
    """

    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    config = {
        "options": {
            "download": {
                "audio_format": "NON-SENS",
            }
        }
    }
    config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "ytpb/config.toml"
    config_path.parent.mkdir(parents=True)
    with config_path.open("w") as f:
        toml.dump(config, f)

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0


@freeze_time("2023-03-26T00:00:00+00:00")
def test_with_config_via_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    """Tests the --config option. The default config file should be ignored."""

    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    default_config = {
        "options": {
            "download": {
                "audio_format": "NON-SENS",
            }
        }
    }
    default_config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "ytpb/config.toml"
    default_config_path.parent.mkdir(parents=True)
    with default_config_path.open("w") as f:
        toml.dump(default_config, f)

    test_config = {
        "options": {
            "download": {
                "audio_format": "itag eq 140",
            }
        }
    }
    test_config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "test-config.toml"
    with test_config_path.open("w") as f:
        toml.dump(test_config, f)

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--config",
                test_config_path,
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-vf",
                "none",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0


@freeze_time("2023-03-26T00:00:00+00:00")
def test_with_non_existent_config_file_via_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    default_config = {
        "options": {
            "download": {
                "audio_format": "NON-SENS",
            }
        }
    }
    default_config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "ytpb/config.toml"
    default_config_path.parent.mkdir(parents=True)
    with default_config_path.open("w") as f:
        toml.dump(default_config, f)

    test_config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "test-config.toml"

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        with pytest.raises(click.FileError) as exc_info:
            ytpb_cli_invoke(
                [
                    "--config",
                    test_config_path,
                    "download",
                    "--no-cache",
                    "--interval",
                    "7959120/7959121",
                    "-vf",
                    "none",
                    stream_url,
                ],
                catch_exceptions=False,
                standalone_mode=False,
            )

    assert str(test_config_path) in str(exc_info)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_no_config_option(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    """Tests the --no-config option. Any configuration files should be ignored.

    In this test, the required -i/--interval option is intentionally
    omitted. Instead, it's defined in the created config file. Since the file
    should be ignored, a click.MissingParameter exception should be raised.
    """

    # Given:
    default_config = {
        "options": {
            "download": {
                "interval": "7959120/7959121",
            }
        }
    }
    default_config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "ytpb/config.toml"
    default_config_path.parent.mkdir(parents=True)
    with default_config_path.open("w") as f:
        toml.dump(default_config, f)

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        with pytest.raises(click.MissingParameter) as exc_info:
            ytpb_cli_invoke(
                [
                    "--no-config",
                    "download",
                    "--no-cache",
                    # (The required -i/--interval option is intentionally omitted.)
                    "-af",
                    "itag eq 140",
                    "-vf",
                    "none",
                    stream_url,
                ],
                catch_exceptions=False,
                standalone_mode=False,
            )

    # Then:
    assert "Missing parameter: interval" in str(exc_info.value)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_conflicting_config_and_no_config_options(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    default_config = {
        "options": {
            "download": {
                "audio_format": "NON-SENS",
            }
        }
    }
    default_config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "ytpb/config.toml"
    default_config_path.parent.mkdir(parents=True)
    with default_config_path.open("w") as f:
        toml.dump(default_config, f)

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        with pytest.raises(click.UsageError) as exc_info:
            ytpb_cli_invoke(
                [
                    "--no-config",
                    "--config",
                    default_config_path,
                    "download",
                    "--no-cache",
                    "-vf",
                    "none",
                    stream_url,
                ],
                catch_exceptions=False,
                standalone_mode=False,
            )

    # Then:
    assert "Conflicting --config and --no-config options given" == str(exc_info.value)


@freeze_time("2023-03-26T00:00:00+00:00")
def test_download_via_stream_id(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    video_id: str,
    audio_base_url: str,
    run_temp_directory: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--no-cut",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                video_id,
            ],
        )

    # Then:
    assert result.exit_code == 0


@freeze_time("2023-03-26T00:00:00+00:00")
def test_download_via_invalid_stream_id(
    ytpb_cli_invoke: Callable,
    run_temp_directory: Path,
) -> None:
    with pytest.raises(click.BadParameter):
        ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "none",
                "INVALID",
            ],
            catch_exceptions=False,
            standalone_mode=False,
        )


@freeze_time("2023-03-26T00:00:00+00:00")
def test_custom_aliases(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    custom_config = {
        "general": {
            "aliases": {
                "custom-alias": "itag eq 140",
            }
        }
    }
    config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "ytpb/config.toml"
    config_path.parent.mkdir(parents=True)
    with config_path.open("w") as f:
        toml.dump(custom_config, f)

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--config",
                config_path,
                "download",
                "--dry-run",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-vf",
                "none",
                "-af",
                "@custom-alias",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0


@freeze_time("2023-03-26T00:00:00+00:00")
def test_dynamic_aliases(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--dry-run",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-vf",
                "none",
                "-af",
                "@140",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "audio_format,video_format",
    [
        ("itag eq 140", "itag eq 0"),
        ("itag eq 0", "itag eq 244"),
    ],
)
@freeze_time("2023-03-26T00:00:00+00:00")
def test_empty_representations(
    audio_format: str,
    video_format: str,
    ytpb_cli_invoke: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "--no-config",
                "download",
                "--dry-run",
                "--no-cut",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                audio_format,
                "-vf",
                video_format,
                stream_url,
            ],
        )

    assert result.exit_code == 1
    assert "error: No streams found matching" in result.output



"""This module contains test cases for the Playback class.

Here are segments and corresponding ingestion start dates and timestamps
used in the tests:
             7959120       21       22
                  |        |        |
 2023-03-25T23:33:54.491Z  56.490Z  58.492Z
                  1679787234.491176
                           1679787236.489910
                                    1679787238.491916
"""

import json
from dataclasses import asdict
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Callable
from urllib.parse import urljoin

import pytest
import responses

from conftest import TEST_DATA_PATH
from freezegun import freeze_time

from ytpb.exceptions import (
    BaseUrlExpiredError,
    CachedItemNotFoundError,
    SequenceLocatingError,
)
from ytpb.fetchers import YtpbInfoFetcher
from ytpb.info import YouTubeVideoInfo
from ytpb.playback import Playback, RewindMoment
from ytpb.streams import AudioOrVideoStream
from ytpb.types import RelativeSegmentSequence


class TestLocateMoment:
    @pytest.fixture(autouse=True)
    def setup_method(
        self,
        fake_info_fetcher: "FakeInfoFetcher",
        add_responses_callback_for_segment_urls: Callable,
        mocked_responses: responses.RequestsMock,
        stream_url: str,
        active_live_video_info: YouTubeVideoInfo,
        audio_base_url: str,
        tmp_path: Path,
    ):
        add_responses_callback_for_segment_urls(urljoin(audio_base_url, r"sq/\w+"))
        self.playback = Playback(stream_url, fetcher=fake_info_fetcher)
        self.playback.fetch_and_set_essential()

    def test_start_sequence(self):
        sequence = 7959120
        date = datetime.fromtimestamp(1679787234.491176, tz=timezone.utc)
        expected = RewindMoment(date, sequence, 0, False)
        assert expected == self.playback.locate_moment(sequence, "140")

    def test_end_sequence(self):
        sequence = 7959120
        date = datetime(2023, 3, 25, 23, 33, 56, 488092, tzinfo=timezone.utc)
        expected = RewindMoment(date, sequence, 0, True)
        assert expected == self.playback.locate_moment(sequence, "140", True)

    def test_start_date(self, add_responses_callback_for_reference_base_url: Callable):
        add_responses_callback_for_reference_base_url()
        date = datetime.fromisoformat("2023-03-25T23:33:55Z")
        expected = RewindMoment(date, 7959120, 0.508824, False)
        assert expected == self.playback.locate_moment(date, "140")

    def test_end_date(self, add_responses_callback_for_reference_base_url: Callable):
        add_responses_callback_for_reference_base_url()
        date = datetime.fromisoformat("2023-03-25T23:33:55Z")
        expected = RewindMoment(date, 7959120, 0.508824, True)
        assert expected == self.playback.locate_moment(date, "140", True)


@pytest.mark.parametrize(
    "start,end",
    [
        (7959120, 7959121),
        (7959120, datetime.fromisoformat("2023-03-25T23:33:57Z")),
        (7959120, timedelta(seconds=3)),  # segment duration (2 s) + 1 s
        (7959120, RelativeSegmentSequence(1)),
        (datetime.fromisoformat("2023-03-25T23:33:55Z"), 7959121),
        (
            datetime.fromisoformat("2023-03-25T23:33:55Z"),
            datetime.fromisoformat("2023-03-25T23:33:57Z"),
        ),
        (datetime.fromisoformat("2023-03-25T23:33:55Z"), timedelta(seconds=2)),
        (datetime.fromisoformat("2023-03-25T23:33:55Z"), RelativeSegmentSequence(1)),
        (timedelta(seconds=3), 7959121),  # segment duration (2 s) + 1 s
        (timedelta(seconds=2), datetime.fromisoformat("2023-03-25T23:33:57Z")),
        (RelativeSegmentSequence(1), 7959121),
        (RelativeSegmentSequence(1), datetime.fromisoformat("2023-03-25T23:33:57Z")),
        (RelativeSegmentSequence(1), 7959121),
    ],
)
def test_locate_interval(
    start,
    end,
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    mocked_responses: responses.RequestsMock,
    stream_url: str,
    active_live_video_info: YouTubeVideoInfo,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    if not (isinstance(start, int) and isinstance(end, int)):
        add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(urljoin(audio_base_url, r"sq/\w+"))

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    interval = playback.locate_interval(start, end, "140")

    # Then:
    assert interval.start.sequence == 7959120
    assert interval.end.sequence == 7959121


@pytest.mark.parametrize(
    "start,end",
    [
        (timedelta(seconds=2), timedelta(seconds=2)),
        (RelativeSegmentSequence(1), RelativeSegmentSequence(1)),
        (timedelta(seconds=2), RelativeSegmentSequence(1)),
        (RelativeSegmentSequence(1), timedelta(seconds=2)),
    ],
)
def test_local_interval_with_relative_start_and_end(
    start: timedelta | RelativeSegmentSequence,
    end: timedelta | RelativeSegmentSequence,
    stream_url: str,
    fake_info_fetcher: "FakeInfoFetcher",
    tmp_path: Path,
):
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    with pytest.raises(ValueError):
        playback.locate_interval(start, end, "140")


@pytest.mark.parametrize(
    "start,end",
    [
        (7959122, 7959121),
        (7959121, datetime.fromisoformat("2023-03-25T23:33:55Z")),
        (datetime.fromisoformat("2023-03-25T23:33:57Z"), 7959120),
        (
            datetime.fromisoformat("2023-03-25T23:33:57Z"),
            datetime.fromisoformat("2023-03-25T23:33:55Z"),
        ),
        (datetime.fromisoformat("2023-03-25T23:33:57Z"), RelativeSegmentSequence(-1)),
    ],
)
def test_locate_interval_with_swapped_start_and_end(
    start,
    end,
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    mocked_responses: responses.RequestsMock,
    stream_url: str,
    active_live_video_info: YouTubeVideoInfo,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    if not (isinstance(start, int) and isinstance(end, int)):
        add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(urljoin(audio_base_url, r"sq/\w+"))

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()

    # Then:
    with pytest.raises((ValueError, SequenceLocatingError)):
        playback.locate_interval(start, end, "140")


def test_create_playback_from_url(
    fake_info_fetcher: "FakeInfoFetcher",
    active_live_video_info: YouTubeVideoInfo,
    streams_in_list: list[AudioOrVideoStream],
    stream_url: str,
    audio_base_url: str,
    run_temp_directory,
):
    playback = Playback.from_url(stream_url, fetcher=fake_info_fetcher)
    assert playback.info == active_live_video_info
    assert len(playback.streams) == len(streams_in_list)


@freeze_time("2023-09-28T17:00:00+00:00")
def test_create_playback_from_manifest(
    fake_info_fetcher: "FakeInfoFetcher",
    active_live_video_info: YouTubeVideoInfo,
    streams_in_list: list[AudioOrVideoStream],
    audio_base_url: str,
    run_temp_directory: Path,
):
    playback = Playback.from_manifest(
        TEST_DATA_PATH / "manifest-1695928670.mpd", fetcher=fake_info_fetcher
    )
    assert playback.info == active_live_video_info
    assert len(playback.streams) == len(streams_in_list)


@freeze_time("2123-09-28T17:00:00+00:00")
def test_create_playback_from_expired_manifest():
    with pytest.raises(BaseUrlExpiredError):
        Playback.from_manifest(TEST_DATA_PATH / "manifest-1695928670.mpd")


@freeze_time("2023-09-28T17:00:00+00:00")
def test_create_playback_from_cache(
    create_cache_file: None,
    stream_url: str,
    active_live_video_info: YouTubeVideoInfo,
    streams_in_list: list[AudioOrVideoStream],
):
    playback = Playback.from_cache(stream_url)
    assert playback.info == active_live_video_info
    assert len(playback.streams) == len(streams_in_list)


@freeze_time("2123-09-28T17:00:00+00:00")
def test_create_playback_from_expired_cache(
    active_live_video_info: YouTubeVideoInfo,
    streams_in_list: list[AudioOrVideoStream],
    video_id: str,
    stream_url: str,
    tmp_path: Path,
):
    # Given:
    test_cache_directory = Playback.get_cache_directory()
    test_cache_directory.mkdir(parents=True)

    frozen_time = datetime.fromisoformat("2123-09-28T17:00:00+00:00").timestamp()
    expired_at = int(frozen_time - 10)
    with open(test_cache_directory / f"{expired_at}~{video_id}", "w") as f:
        test_cache = {
            "info": asdict(active_live_video_info),
            "streams": [asdict(stream) for stream in streams_in_list],
        }
        json.dump(test_cache, f)

    # When:
    with pytest.raises(CachedItemNotFoundError):
        Playback.from_cache(stream_url)


def test_create_playback_from_not_found_cache(stream_url: str, tmp_path: Path):
    with pytest.raises(CachedItemNotFoundError):
        Playback.from_cache(stream_url)


def test_type_of_playback_default_fetcher(stream_url: str):
    playback = Playback(stream_url)
    assert isinstance(playback.fetcher, YtpbInfoFetcher)



"""A naive disk-based caching solution via JSON. Invalidation is based on the
filenaming."""

import json
import time
from pathlib import Path
from typing import Iterable

import structlog

logger = structlog.get_logger(__name__)


def _find_cached_item_paths(key: str, cache_directory: Path) -> Iterable[Path]:
    return cache_directory.glob(f"*~{key}")


def _check_item_is_expired(item_name: str) -> bool:
    """Check if an item has expired or not based on the item filename."""
    expires_at = int(item_name.split("~")[0])
    return time.time() >= expires_at


def read_from_cache(key: str, cache_directory: Path) -> dict | None:
    try:
        found_item_paths = _find_cached_item_paths(key, cache_directory)
        *earlier_item_paths, latest_item_path = sorted(found_item_paths)
    except ValueError:
        item = None
    else:
        for path in earlier_item_paths:
            path.unlink()

        if _check_item_is_expired(latest_item_path.name):
            logger.debug("Found expired cached item: %s", latest_item_path)
            latest_item_path.unlink()
            item = None
        else:
            with open(latest_item_path) as f:
                item = json.load(f)
            logger.debug("Found unexpired cached item: %s", latest_item_path)

    return item


def write_to_cache(
    key: str, expires_at: str, item: dict, cache_directory: Path
) -> None:
    """Write an item to the cache file. The previous cached items with the key
    will be removed before."""
    cache_directory.mkdir(parents=True, exist_ok=True)
    if old_item_paths := _find_cached_item_paths(key, cache_directory):
        for path in old_item_paths:
            path.unlink()
    new_item_path = cache_directory / f"{expires_at}~{key}"
    with open(new_item_path, "w") as f:
        json.dump(item, f)
    logger.debug("New cache item has been created: %s", new_item_path)


def remove_expired_cache_items(cache_directory: Path) -> None:
    for path in sorted(cache_directory.glob("*~*")):
        if _check_item_is_expired(path.name):
            path.unlink()
        else:
            break



import atexit
import logging
import os
import re
import sys
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timezone
from fileinput import FileInput
from pathlib import Path
from typing import Any, cast, TextIO

import click
import cloup
import structlog
import toml

from ytpb.cli.commands.capture import capture_group
from ytpb.cli.commands.download import download_command
from ytpb.cli.commands.mpd import mpd_group
from ytpb.cli.options import config_options, logging_options
from ytpb.config import (
    ALL_ALIASES,
    DEFAULT_CONFIG,
    get_default_config_path,
    load_config_from_file,
    setup_logging,
    update_nested_dict,
)
from ytpb.types import ConfigMap

logger = structlog.get_logger(__name__)


@dataclass
class ReportStreamWrapper:
    stream: TextIO
    file: TextIO

    def write(self, message: str) -> None:
        self.stream.write(message)
        self.file.write(message)

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.stream, attr)


@dataclass
class ContextObject:
    """This object is referenced as `ctx.obj`."""

    config = ConfigMap(DEFAULT_CONFIG)


def load_config_into_context(ctx: click.Context, path: Path) -> dict:
    try:
        config_dict = load_config_from_file(path)
    except FileNotFoundError as e:
        raise click.FileError(
            filename=path,
            hint=str(e),
        )
    except toml.TomlDecodeError as e:
        raise click.FileError(
            filename=path,
            hint=f"Could not load configuration file.\n{e}",
        )

    ctx.ensure_object(ContextObject)
    ctx.obj.config = ctx.obj.config.new_child(config_dict)

    default_map_from_config = ctx.obj.config["options"]
    ctx.default_map = update_nested_dict(ctx.default_map, default_map_from_config)

    try:
        ALL_ALIASES.update(ctx.obj.config["general"]["aliases"])
    except KeyError:
        pass


@cloup.group(invoke_without_command=True)
@cloup.option_group(
    "Global options",
    config_options,
    logging_options,
)
@click.pass_context
def base_cli(
    ctx: click.Context, config_path: Path, no_config: bool, debug: bool, report: bool
) -> None:
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        ctx.exit()

    ctx.ensure_object(ContextObject)
    ctx.default_map = ctx.obj.config["options"]

    if report:
        debug = True
        os.environ["NO_COLOR"] = "1"

        timestamp = datetime.now(tz=timezone.utc).strftime("%Y%m%d-%H%M%S")
        report_handle = open(Path.cwd() / f"ytpb-{timestamp}.log", "a")
        sys.stdout = cast(TextIO, ReportStreamWrapper(sys.stdout, report_handle))
        sys.stderr = cast(TextIO, ReportStreamWrapper(sys.stderr, report_handle))

        @atexit.register
        def sanitize_report_file() -> None:
            report_handle.close()
            with FileInput(report_handle.name, inplace=True) as fi:
                for line in fi:
                    print(re.sub("/ip/([\w.:]+)/", "/ip/0.0.0.0/", line), end="")

    setup_logging(logging.DEBUG if debug else logging.WARNING)

    if not no_config:
        if config_path:
            logger.info("Using configuration from '%s' via --config", config_path)
            load_config_into_context(ctx, config_path)
        else:
            default_config_path = get_default_config_path()
            if default_config_path.exists():
                logger.info("Using configuration from '%s'", default_config_path)
                load_config_into_context(ctx, default_config_path)
    else:
        if config_path:
            raise click.UsageError("Conflicting --config and --no-config options given")


cli = deepcopy(base_cli)

cli.help = "A playback for YouTube live streams"
cli.section(
    "Top-level commands",
    download_command,
    capture_group,
    mpd_group,
)



from datetime import datetime, timedelta, timezone

import pytest

from ytpb.utils.date import (
    build_style_parameters_from_spec,
    DurationFormatPattern,
    express_timedelta_in_words,
    format_duration,
    format_iso_datetime,
    format_timedelta,
    ISO8601DateStyleParameters,
)


class TestFormatDatetime:
    @pytest.mark.parametrize(
        "date,expected",
        [
            ("2023-08-09T10+01:00", "2023-08-09T10:00:00+01"),
            ("2023-08-09T10:20+01:00", "2023-08-09T10:20:00+01"),
            ("2023-08-09T10:20:30+01:00", "2023-08-09T10:20:30+01"),
            ("2023-08-09T10:00:00.123+01:00", "2023-08-09T10:00:00+01"),
        ],
    )
    def test_extended_format_and_complete_precision(self, date, expected):
        assert expected == format_iso_datetime(
            datetime.fromisoformat(date),
            ISO8601DateStyleParameters(format="extended", precision="complete"),
        )

    @pytest.mark.parametrize(
        "date,expected",
        [
            ("2023-08-09T00+01:00", "2023-08-09T00+01"),
            ("2023-08-09T10+01:00", "2023-08-09T10+01"),
            ("2023-08-09T10:20+01:00", "2023-08-09T10:20+01"),
            ("2023-08-09T10:20:30+01:00", "2023-08-09T10:20:30+01"),
            ("2023-08-09T10:00:00.123+01:00", "2023-08-09T10:00:00+01"),
        ],
    )
    def test_extended_format_and_reduced_precision(self, date, expected):
        assert expected == format_iso_datetime(
            datetime.fromisoformat(date),
            ISO8601DateStyleParameters(format="extended", precision="reduced"),
        )

    @pytest.mark.parametrize(
        "date,expected",
        [
            ("2023-08-09T10+01:00", "20230809T100000+01"),
            ("2023-08-09T10:20+01:00", "20230809T102000+01"),
            ("2023-08-09T10:20:30+01:00", "20230809T102030+01"),
            ("2023-08-09T10:00:00.123+01:00", "20230809T100000+01"),
        ],
    )
    def test_basic_format_and_complete_precision(self, date, expected):
        assert expected == format_iso_datetime(
            datetime.fromisoformat(date),
            ISO8601DateStyleParameters(format="basic", precision="complete"),
        )

    @pytest.mark.parametrize(
        "date,expected",
        [
            ("2023-08-09T10+01:00", "20230809T10+01"),
            ("2023-08-09T10:20+01:00", "20230809T1020+01"),
            ("2023-08-09T10:20:30+01:00", "20230809T102030+01"),
            ("2023-08-09T10:00:00.123+01:00", "20230809T100000+01"),
        ],
    )
    def test_basic_format_and_reduced_precision(self, date, expected):
        assert expected == format_iso_datetime(
            datetime.fromisoformat(date),
            ISO8601DateStyleParameters(format="basic", precision="reduced"),
        )

    @pytest.mark.parametrize(
        "basic_or_extended,offset_format,expected",
        [
            ("basic", "hh", "20230809T102030+01"),
            ("basic", "hhmm", "20230809T102030+0100"),
            ("extended", "hh", "2023-08-09T10:20:30+01"),
            ("extended", "hhmm", "2023-08-09T10:20:30+01:00"),
        ],
    )
    def test_utc_offset_formats(self, basic_or_extended, offset_format, expected):
        assert expected == format_iso_datetime(
            datetime.fromisoformat("2023-08-09T10:20:30+01:00"),
            ISO8601DateStyleParameters(
                format=basic_or_extended, offset_format=offset_format
            ),
        )

    def test_use_z_for_utc(self):
        assert "20230809T102030Z" == format_iso_datetime(
            datetime.fromisoformat("2023-08-09T10:20:30+00:00"),
            ISO8601DateStyleParameters(use_z_for_utc=True),
        )
        assert "20230809T102030+01" == format_iso_datetime(
            datetime.fromisoformat("2023-08-09T10:20:30+01:00"),
            ISO8601DateStyleParameters(use_z_for_utc=True),
        )
        assert "20230809T102030+00" == format_iso_datetime(
            datetime.fromisoformat("2023-08-09T10:20:30+00:00"),
            ISO8601DateStyleParameters(use_z_for_utc=False),
        )

    def test_raise_with_naive_datetime(self):
        with pytest.raises(ValueError):
            format_iso_datetime(
                datetime.now(),
            )

    def test_raise_with_invalid_argument(self):
        with pytest.raises(ValueError):
            format_iso_datetime(
                datetime.now(timezone.utc), ISO8601DateStyleParameters(format="invalid")
            )
        with pytest.raises(ValueError):
            format_iso_datetime(
                datetime.now(timezone.utc),
                ISO8601DateStyleParameters(precision="invalid"),
            )
        with pytest.raises(ValueError):
            format_iso_datetime(
                datetime.now(timezone.utc),
                ISO8601DateStyleParameters(offset_format="invalid"),
            )


@pytest.mark.parametrize(
    "delta,expected",
    [
        (timedelta(hours=1), "PT1H"),
        (timedelta(hours=1, minutes=2), "PT1H2M"),
        (timedelta(hours=1, minutes=2, seconds=3), "PT1H2M3S"),
        (timedelta(hours=0, minutes=2, seconds=3), "PT2M3S"),
        (timedelta(hours=0, minutes=0, seconds=3), "PT3S"),
        (timedelta(hours=1, minutes=0, seconds=3), "PT1H3S"),
    ],
)
def test_format_duration(delta, expected: str):
    assert expected == format_duration(delta, DurationFormatPattern.ISO8601)


@pytest.mark.parametrize(
    "delta,expected",
    [
        (timedelta(hours=1), "an hour"),
        (timedelta(hours=1, minutes=14), "an hour"),
        (timedelta(hours=1, minutes=15), "1.5 hours"),
        (timedelta(hours=1, minutes=44), "1.5 hours"),
        (timedelta(hours=1, minutes=45), "2 hours"),
        (timedelta(hours=1, minutes=44, seconds=30), "2 hours"),
        (timedelta(minutes=1), "a minute"),
        (timedelta(minutes=45), "an hour"),
        (timedelta(minutes=44), "44 minutes"),
        (timedelta(minutes=44, seconds=29), "44 minutes"),
        (timedelta(minutes=44, seconds=30), "an hour"),
        (timedelta(seconds=45), "less than a minute"),
        (timedelta(seconds=10), "a moment"),
    ],
)
def test_express_timedelta_in_words(delta, expected):
    assert expected == express_timedelta_in_words(delta)


@pytest.mark.parametrize(
    "duration,expected",
    [
        (timedelta(seconds=0), "+00:00"),
        (timedelta(seconds=1.123), "+00:01"),
        (timedelta(seconds=-1.123), "-00:01"),
        (timedelta(seconds=121.100), "+02:01"),
        (timedelta(hours=1, seconds=3), "+01:00:03"),
    ],
)
def test_format_timedelta(duration, expected):
    result = format_timedelta(duration, use_ms_precision=False)
    assert result == expected


@pytest.mark.parametrize(
    "duration,expected",
    [
        (timedelta(seconds=0), "+00:00.000"),
        (timedelta(seconds=0.1), "+00:00.100"),
        (timedelta(seconds=1.123789), "+00:01.124"),
    ],
)
def test_format_timedelta_with_ms_precision(duration, expected):
    result = format_timedelta(duration, use_ms_precision=True)
    assert result == expected


class TestBuildStyleParametersFromSpec:
    def test_parse_single_date_style(self):
        expected = ISO8601DateStyleParameters(
            format="extended",
        )
        assert expected == build_style_parameters_from_spec("extended")

    @pytest.mark.parametrize("spec", ["extended,complete", "extended, complete"])
    def test_parse_date_styles(self, spec):
        expected = ISO8601DateStyleParameters(
            format="extended",
            precision="complete",
        )
        assert expected == build_style_parameters_from_spec(spec)

    def test_mutually_exlusive_styles(self):
        with pytest.raises(ValueError) as exc_info:
            build_style_parameters_from_spec("basic,extended")
        message = "Mutually exclusive styles provided: 'basic' and 'extended'"
        assert message == str(exc_info.value)

    def test_unknown_styles(self):
        with pytest.warns() as record:
            build_style_parameters_from_spec("unknown")
        assert str(record[0].message) == "Ignoring unknown style(s): unknown"

        with pytest.warns() as record:
            build_style_parameters_from_spec("unknown,invalid")
        assert str(record[0].message) == "Ignoring unknown style(s): invalid, unknown"

    def test_empty_or_none_style(self):
        assert None is build_style_parameters_from_spec("")
        assert None is build_style_parameters_from_spec(None)



import pytest

from ytpb.exceptions import BadCommandArgument
from ytpb.utils.url import (
    extract_media_type_from_url,
    extract_parameter_from_url,
    normalize_video_url,
)


def test_extract_itag_from_url(audio_base_url):
    assert extract_parameter_from_url("itag", audio_base_url) == "140"


def test_extract_mime_type_from_url(audio_base_url):
    assert extract_media_type_from_url(audio_base_url) == ("audio", "mp4")


@pytest.mark.parametrize(
    "stream_url_or_id",
    [
        "https://www.youtube.com/watch?v=kHwmzef842g",
        "https://www.youtube.com/watch?v=kHwmzef842g&param=value",
        "https://www.youtu.be/kHwmzef842g",
        "kHwmzef842g",
    ],
)
def test_normalize_video_url(stream_url_or_id):
    expected = "https://www.youtube.com/watch?v=kHwmzef842g"
    assert expected == normalize_video_url(stream_url_or_id)


@pytest.mark.parametrize(
    "stream_url_or_id",
    [
        "https://www.youtube.com/watch?v=kHwmzef842gxxx",
        "https://www.youtube.com/watch?v=kHwmzef",
        "https://www.youtu.be/kHwmzef842gxxx",
        "kHwmzef842gxxx",
        "kHwmzef",
    ],
)
def test_failed_normalize_video_url(stream_url_or_id):
    with pytest.raises(BadCommandArgument):
        normalize_video_url(stream_url_or_id)



from datetime import datetime, time, timedelta
from enum import auto, StrEnum
from typing import Literal, NamedTuple

import click
import structlog
from timedelta_isoformat import timedelta as isotimedelta

from ytpb import types
from ytpb.conditional import FORMAT_SPEC_RE
from ytpb.types import SegmentSequence
from ytpb.utils.date import ensure_date_aware

logger = structlog.get_logger(__name__)


class PointInStreamParamType(click.ParamType):
    def __init__(self, allowed_literals: list[str] = None):
        if allowed_literals:
            self.allowed_literals = allowed_literals
        else:
            self.allowed_literals = []

    def convert(self, value: str, param, ctx) -> types.AbsolutePointInStream | str:
        match value:
            # Allowed literals
            case literal if value in self.allowed_literals:
                output = literal
            # Time of today
            case x if x[0] == "T" or (":" in x and "-" not in x):
                parsed_time = time.fromisoformat(x)
                today = datetime.now()
                output = today.replace(
                    hour=parsed_time.hour,
                    minute=parsed_time.minute,
                    second=parsed_time.second,
                    microsecond=parsed_time.microsecond,
                )
                output = output.astimezone(parsed_time.tzinfo)
            # Segment sequence
            case sequence if value.isdecimal():
                output = SegmentSequence(sequence)
            # Seems like a date and time
            case value if value[:4].isdecimal():
                try:
                    output = ensure_date_aware(datetime.fromisoformat(value))
                except ValueError:
                    message = f"'{value}' does not match ISO 8601 date format."
                    self.fail(message, param, ctx)
            case _:
                self.fail("Option doesn't allow '{}' value", param, ctx)
        return output


class InputRewindInterval(NamedTuple):
    start: SegmentSequence | datetime | timedelta
    end: SegmentSequence | datetime | timedelta | Literal["now"]


class RewindIntervalParamType(click.ParamType):
    error_message_fmt = "Incorrectly formated interval: {}"

    def _replace_datetime_components(self, date: datetime, value: str):
        date_units = {"Y": "year", "M": "month", "D": "day"}
        time_units = {"H": "hour", "M": "minute", "S": "second"}

        components_to_replace = {}

        for part, units in zip(value.split("T"), (date_units, time_units)):
            if not part:
                continue
            previous = 0
            for i, c in enumerate(part):
                if c in units:
                    components_to_replace[units[c]] = int(part[previous:i])
                    previous = i + 1

        return date.replace(**components_to_replace)

    def _parse_interval_part(
        self,
        part: str,
        end: bool = False,
    ) -> int | str | Literal["now", ".."] | datetime | timedelta | isotimedelta:
        match part:
            # Sequence number
            case x if x.isdecimal():
                output = int(x)
            # Duration
            case x if x[0] == "P":
                output = isotimedelta.fromisoformat(x)
            # Replacing components
            case x if set(x) & set("DHMS"):
                output = x
            # Time of today
            case x if x[0] == "T" or (":" in x and "-" not in x):
                parsed_time = time.fromisoformat(x)
                today = datetime.now()
                output = today.replace(
                    hour=parsed_time.hour,
                    minute=parsed_time.minute,
                    second=parsed_time.second,
                    microsecond=parsed_time.microsecond,
                )
                output = output.astimezone(parsed_time.tzinfo)
            # Date and time
            case x if "T" in x:
                output = datetime.fromisoformat(x)
            case "now" | ".." as x:
                output = x
            case _:
                raise click.BadParameter(f"Incorrectly formatted part: {part}")

        return output

    def convert(
        self, value: str, param: click.Parameter, ctx: click.Context
    ) -> InputRewindInterval:
        parts = value.split("/")
        if len(parts) != 2:
            raise click.BadParameter("Interval must be formatted as '<start>/<end>'")

        start_part, end_part = parts

        try:
            parsed_start = self._parse_interval_part(start_part)
        except ValueError as exc:
            raise click.BadParameter(f"'{start_part}', {exc}.")

        try:
            parsed_end = self._parse_interval_part(end_part)
        except ValueError as exc:
            raise click.BadParameter(f"'{end_part}', {exc}.")

        match parsed_start, parsed_end:
            # Two durations
            case [timedelta(), timedelta()]:
                raise click.BadParameter("Two durations are ambiguous.")
            case ["now" | ".." as x, _]:
                raise click.BadParameter(
                    f"Keyword '{x}' is only allowed for the end part."
                )
            # Duration and '..'
            case [timedelta(), ".."]:
                raise click.BadParameter("Keyword '..' not compatible with duration.")
            # Anything compatible and 'now' or '..'
            case [int() | datetime() | timedelta(), "now" | ".."]:
                start = parsed_start
                end = parsed_end
            # Replacing components and date and time
            case (str(), datetime()) | (datetime(), str()):
                if isinstance(parsed_start, str):
                    end = parsed_end
                    start = self._replace_datetime_components(end, parsed_start)
                else:
                    start = parsed_start
                    end = self._replace_datetime_components(start, parsed_end)
            # Replacing components and anything remaining, non-compatible
            case (str(), _) | (_, str()):
                raise click.BadParameter(
                    "Replacement components is only compatible with date and time."
                )
            # Remaining and compatible
            case [_, _]:
                start = parsed_start
                end = parsed_end

        if type(start) == type(end) and start >= end:
            raise click.BadParameter(
                f"Start is ahead or equal to end: {start} >= {end}"
            )

        return start, end


class FormatSpecType(StrEnum):
    AUDIO = auto()
    VIDEO = auto()


class FormatSpecParamType(click.ParamType):
    def __init__(self, format_spec_type: FormatSpecType):
        self.format_spec_type = format_spec_type

    def convert(
        self, value: str, param: click.Parameter, ctx: click.Context
    ) -> str | None:
        if value in ("", "none"):
            output = None
        else:
            guard_condition = f"mime_type contains {self.format_spec_type.value}"
            if matched := FORMAT_SPEC_RE.match(value):
                if matched.group("function"):
                    replace_with = f"{guard_condition} and [{matched.group('expr')}]"
                    output = value.replace(matched.group("expr"), replace_with)
                else:
                    output = f"{guard_condition} and [{value}]"

        return output



import os
from pathlib import Path
from typing import Callable
from unittest.mock import MagicMock, patch
from urllib.parse import urljoin

import pytest

import toml
from freezegun import freeze_time


@pytest.mark.parametrize(
    "audio_formats,video_formats",
    [
        ("itag eq 140", "itag eq 243 or itag eq 244"),
        ("itag eq 140", "none"),
        ("none", "itag eq 243 or itag eq 244"),
    ],
)
def test_compose_mpd(
    audio_formats,
    video_formats,
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    tmp_path: Path,
    expected_out,
) -> None:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                audio_formats,
                "-vf",
                video_formats,
                stream_url,
            ],
        )
    assert result.exit_code == 0
    assert result.output == expected_out


@freeze_time("2023-03-26T00:00:00+00:00")
def test_compose_mpd_with_no_streams(
    ytpb_cli_invoke: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_id: str,
    audio_base_url: str,
    video_base_url: str,
    tmp_path: Path,
    expected_out,
) -> None:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "2023-03-25T23:33:55+00:00/2023-03-25T23:33:58+00:00",
                "-af",
                "itag eq 0",
                "-vf",
                "itag eq 0",
                stream_url,
            ],
        )

    assert result.exit_code == 1
    assert result.output == expected_out


def test_compose_mpd_using_yt_dlp(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_id: str,
    audio_base_url: str,
    video_base_url: str,
    tmp_path: Path,
) -> None:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    with patch("ytpb.cli.common.YoutubeDLInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--yt-dlp",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "",
                stream_url,
            ],
        )

    assert result.exit_code == 0


def test_with_default_config(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )

    config = {
        "options": {
            "mpd": {
                "compose": {
                    "audio_formats": "NON-SENS",
                }
            }
        }
    }
    config_path = Path(os.getenv("XDG_CONFIG_HOME")) / "ytpb/config.toml"
    config_path.parent.mkdir(parents=True)
    with config_path.open("w") as f:
        toml.dump(config, f)

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "itag eq 243",
                stream_url,
            ],
        )

    # Then:
    assert result.exit_code == 0


def test_compose_to_custom_output_path(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    tmp_path: Path,
):
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "itag eq 243",
                "-o",
                "manifest.xml",
                stream_url,
            ],
            catch_exceptions=False,
        )
    assert result.exit_code == 0
    assert os.path.exists("manifest.xml")


def test_compose_to_custom_template_output_path(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_id: str,
    tmp_path: Path,
):
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "itag eq 243",
                "-o",
                "<id>.xml",
                stream_url,
            ],
            catch_exceptions=False,
        )
    assert result.exit_code == 0
    assert os.path.exists(f"{video_id}.xml")



import os
from pathlib import Path

import pytest

from conftest import TEST_DATA_PATH

from helpers import assert_approx_duration, assert_number_of_streams
from ytpb.merge import merge_segments

SEGMENT_BASE_PATH = Path(TEST_DATA_PATH) / "segments"


def setup_function():
    os.environ["YTPB_VP9_ENCODING_SETTINGS"] = (
        "libvpx-vp9 -crf 63 -b:v 0 -deadline realtime -cpu-used 8 -row-mt 1 -s 1x1"
    )


def test_merge_with_default_encoding_settings(run_temp_directory: Path, tmp_path: Path):
    """Test with default video encoding settings."""
    os.environ["YTPB_VP9_ENCODING_SETTINGS"] = ""

    output_stem = tmp_path / "merged"

    merge_segments(
        audio_segment_paths=[SEGMENT_BASE_PATH / "7959120.i140.mp4"],
        video_segment_paths=[SEGMENT_BASE_PATH / "7959120.i244.webm"],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
        cut_at_start=500,
    )

    expected_output_path = output_stem.with_suffix(".mkv")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 2)
    assert_approx_duration(expected_output_path, 1.5)

    assert not any(run_temp_directory.iterdir())


def test_merge_without_cleaning_up(run_temp_directory: Path, tmp_path: Path):
    output_stem = tmp_path / "merged"

    merge_segments(
        audio_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i140.mp4",
            SEGMENT_BASE_PATH / "7959121.i140.mp4",
        ],
        video_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i244.webm",
            SEGMENT_BASE_PATH / "7959121.i244.webm",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
        cleanup=False,
    )

    assert any(run_temp_directory.iterdir())


def test_merge_one_audio_and_video_segments_without_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"

    merge_segments(
        audio_segment_paths=[SEGMENT_BASE_PATH / "7959120.i140.mp4"],
        video_segment_paths=[SEGMENT_BASE_PATH / "7959120.i244.webm"],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
    )

    expected_output_path = output_stem.with_suffix(".mkv")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 2)
    assert_approx_duration(expected_output_path, 2.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_one_audio_and_video_segments_with_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"

    merge_segments(
        audio_segment_paths=[SEGMENT_BASE_PATH / "7959120.i140.mp4"],
        video_segment_paths=[SEGMENT_BASE_PATH / "7959120.i244.webm"],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
        cut_at_start=500,
        cut_at_end=500,
    )

    expected_output_path = output_stem.with_suffix(".mkv")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 2)
    assert_approx_duration(expected_output_path, 1.5)

    assert not any(run_temp_directory.iterdir())


def test_merge_two_audio_and_video_segments_without_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"

    merge_segments(
        audio_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i140.mp4",
            SEGMENT_BASE_PATH / "7959121.i140.mp4",
        ],
        video_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i244.webm",
            SEGMENT_BASE_PATH / "7959121.i244.webm",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
    )

    expected_output_path = output_stem.with_suffix(".mkv")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 2)
    assert_approx_duration(expected_output_path, 4.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_two_audio_and_video_segments_with_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"

    merge_segments(
        audio_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i140.mp4",
            SEGMENT_BASE_PATH / "7959121.i140.mp4",
        ],
        video_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i244.webm",
            SEGMENT_BASE_PATH / "7959121.i244.webm",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
        cut_at_start=500,
        cut_at_end=500,
    )

    expected_output_path = output_stem.with_suffix(".mkv")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 2)
    assert_approx_duration(expected_output_path, 3.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_three_audio_and_video_segments_without_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"
    merge_segments(
        audio_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i140.mp4",
            SEGMENT_BASE_PATH / "7959121.i140.mp4",
            SEGMENT_BASE_PATH / "7959122.i140.mp4",
        ],
        video_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i244.webm",
            SEGMENT_BASE_PATH / "7959121.i244.webm",
            SEGMENT_BASE_PATH / "7959122.i244.webm",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
    )

    expected_output_path = output_stem.with_suffix(".mkv")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 2)
    assert_approx_duration(expected_output_path, 6.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_three_audio_and_video_segments_with_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"
    merge_segments(
        audio_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i140.mp4",
            SEGMENT_BASE_PATH / "7959121.i140.mp4",
            SEGMENT_BASE_PATH / "7959122.i140.mp4",
        ],
        video_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i244.webm",
            SEGMENT_BASE_PATH / "7959121.i244.webm",
            SEGMENT_BASE_PATH / "7959122.i244.webm",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
        cut_at_start=500,
        cut_at_end=500,
    )

    expected_output_path = output_stem.with_suffix(".mkv")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 2)
    assert_approx_duration(expected_output_path, 5.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_two_audio_segments_without_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"
    merge_segments(
        audio_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i140.mp4",
            SEGMENT_BASE_PATH / "7959121.i140.mp4",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
    )

    expected_output_path = output_stem.with_suffix(".mp4")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 1)
    assert_approx_duration(expected_output_path, 4.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_two_audio_segments_with_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"
    merge_segments(
        audio_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i140.mp4",
            SEGMENT_BASE_PATH / "7959121.i140.mp4",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
        cut_at_start=500,
        cut_at_end=500,
    )

    expected_output_path = output_stem.with_suffix(".mp4")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 1)
    assert_approx_duration(expected_output_path, 3.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_two_video_segments_without_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"
    merge_segments(
        video_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i244.webm",
            SEGMENT_BASE_PATH / "7959121.i244.webm",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
    )

    expected_output_path = output_stem.with_suffix(".webm")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 1)
    assert_approx_duration(expected_output_path, 4.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_two_video_segments_with_cutting(
    run_temp_directory: Path, tmp_path: Path
):
    output_stem = tmp_path / "merged"
    merge_segments(
        video_segment_paths=[
            SEGMENT_BASE_PATH / "7959120.i244.webm",
            SEGMENT_BASE_PATH / "7959121.i244.webm",
        ],
        output_stem=output_stem,
        temp_directory=run_temp_directory,
        cut_at_start=500,
        cut_at_end=500,
    )

    expected_output_path = output_stem.with_suffix(".webm")
    assert os.path.exists(expected_output_path)

    assert_number_of_streams(expected_output_path, 1)
    assert_approx_duration(expected_output_path, 3.0)

    assert not any(run_temp_directory.iterdir())


def test_merge_unequal_number_of_audio_and_video_segments(
    run_temp_directory: Path, tmp_path: Path
):
    with pytest.raises(ValueError) as e:
        merge_segments(
            audio_segment_paths=[
                SEGMENT_BASE_PATH / "7959120.i140.mp4",
                SEGMENT_BASE_PATH / "7959121.i140.mp4",
            ],
            video_segment_paths=[
                SEGMENT_BASE_PATH / "7959120.i244.webm",
            ],
            output_stem=tmp_path / "merged",
            temp_directory=run_temp_directory,
        )

    assert "of equal size" in str(e.value)



import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Callable
from urllib.parse import urljoin

import responses

from ytpb import actions
from ytpb.playback import Playback


@dataclass
class FakeRewindMoment:
    sequence: int


@dataclass
class FakeRewindInterval:
    start: FakeRewindMoment
    end: FakeRewindMoment


def test_download_excerpt_between_sequences_without_merging(
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_segment_urls: Callable,
    stream_url: str,
    audio_base_url: str,
    video_base_url: str,
    run_temp_directory: Path,
) -> None:
    # Given:
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"), urljoin(video_base_url, r"sq/\w+")
    )

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    _, *output_paths = actions.download.download_excerpt(
        playback,
        FakeRewindInterval(FakeRewindMoment(7959120), FakeRewindMoment(7959121)),
        "itag eq 140",
        "itag eq 244",
        no_merge=True,
    )

    # Then:
    assert output_paths == [
        None,
        [
            run_temp_directory / "segments/7959120.i140.mp4",
            run_temp_directory / "segments/7959121.i140.mp4",
        ],
        [
            run_temp_directory / "segments/7959120.i244.webm",
            run_temp_directory / "segments/7959121.i244.webm",
        ],
    ]
    assert os.path.exists(run_temp_directory / "segments/7959120.i140.mp4")
    assert os.path.exists(run_temp_directory / "segments/7959121.i140.mp4")
    assert os.path.exists(run_temp_directory / "segments/7959120.i244.webm")
    assert os.path.exists(run_temp_directory / "segments/7959121.i244.webm")


def test_download_audio_excerpt_between_sequences_without_merging(
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_segment_urls: Callable,
    stream_url: str,
    audio_base_url: str,
    run_temp_directory: Path,
) -> None:
    # Given:
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    _, *output_paths = actions.download.download_excerpt(
        playback,
        FakeRewindInterval(FakeRewindMoment(7959120), FakeRewindMoment(7959121)),
        "itag eq 140",
        no_merge=True,
    )

    # Then:
    assert output_paths == [
        None,
        [
            run_temp_directory / "segments/7959120.i140.mp4",
            run_temp_directory / "segments/7959121.i140.mp4",
        ],
        [],
    ]
    assert os.path.exists(run_temp_directory / "segments/7959120.i140.mp4")
    assert os.path.exists(run_temp_directory / "segments/7959121.i140.mp4")


def test_download_excerpt_between_dates_without_merging(
    fake_info_fetcher: "FakeInfoFetcher",
    mocked_responses: responses.RequestsMock,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    stream_url: str,
    audio_base_url: str,
    video_base_url: str,
    run_temp_directory: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"), urljoin(video_base_url, r"sq/\w+")
    )

    # When:
    start_date = datetime.fromisoformat("2023-03-25T23:33:55+00:00")
    end_date = datetime.fromisoformat("2023-03-25T23:33:57+00:00")

    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()

    rewind_interval = playback.locate_interval(start_date, end_date, itag="140")
    _, *output_paths = actions.download.download_excerpt(
        playback, rewind_interval, "itag eq 140", "itag eq 244", no_merge=True
    )

    # Then:
    assert output_paths == [
        None,
        [
            run_temp_directory / "segments/7959120.i140.mp4",
            run_temp_directory / "segments/7959121.i140.mp4",
        ],
        [
            run_temp_directory / "segments/7959120.i244.webm",
            run_temp_directory / "segments/7959121.i244.webm",
        ],
    ]

    assert os.path.exists(run_temp_directory / "segments/7959120.i140.mp4")
    assert os.path.exists(run_temp_directory / "segments/7959121.i140.mp4")
    assert os.path.exists(run_temp_directory / "segments/7959120.i244.webm")
    assert os.path.exists(run_temp_directory / "segments/7959121.i244.webm")



from dataclasses import asdict

from conftest import TEST_DATA_PATH

from ytpb import info


def test_extract_video_info(stream_url: str, active_live_video_info):
    with open(TEST_DATA_PATH / "webpage-1695928670.html") as f:
        result = info.extract_video_info(stream_url, f.read())
    assert asdict(result) == asdict(active_live_video_info)



from pathlib import Path
from unittest.mock import MagicMock, patch

import responses
from yt_dlp import YoutubeDL

from ytpb.fetchers import YoutubeDLInfoFetcher, YtpbInfoFetcher
from ytpb.info import YouTubeVideoInfo
from ytpb.playback import Playback
from ytpb.streams import AudioOrVideoStream


@patch("ytpb.fetchers.extract_video_info")
@patch("ytpb.fetchers.extract_representations_info")
def test_ytpb_info_fetcher(
    mock_extract_representations_info: MagicMock,
    mock_extract_video_info: MagicMock,
    mocked_responses: responses.RequestsMock,
    stream_url: str,
    active_live_video_info: YouTubeVideoInfo,
    streams_in_list: list[AudioOrVideoStream],
    tmp_path: Path,
):
    mocked_responses.get(stream_url, status=200)
    mocked_responses.get(active_live_video_info.dash_manifest_url, status=200)

    mock_extract_video_info.return_value = active_live_video_info
    mock_extract_representations_info.return_value = streams_in_list

    fetcher = YtpbInfoFetcher(stream_url)
    assert fetcher.fetch_video_info() == active_live_video_info
    assert len(fetcher.fetch_streams()) == len(streams_in_list)


def test_youtube_dl_info_fetcher(
    stream_url: str,
    active_live_video_info: YouTubeVideoInfo,
    streams_in_list: list[AudioOrVideoStream],
    youtube_dl_info: dict,
    tmp_path: Path,
):
    with patch.object(YoutubeDL, "extract_info", autospec=True) as mock:
        mock.return_value = youtube_dl_info

        fetcher = YoutubeDLInfoFetcher(stream_url)
        playback = Playback(stream_url, fetcher=fetcher)
        playback.fetch_and_set_essential()

    assert playback.info == active_live_video_info
    assert len(playback.streams) == len(streams_in_list)



import copy
from datetime import datetime, timezone

from lxml import etree
from lxml.builder import E

from ytpb.exceptions import YtpbError
from ytpb.mpd import NAMESPACES as NS
from ytpb.playback import Playback, RewindInterval
from ytpb.segment import SegmentMetadata
from ytpb.streams import SetOfStreams
from ytpb.utils.other import S_TO_MS
from ytpb.utils.url import extract_parameter_from_url


def _build_top_level_comment_text(base_url: str) -> str:
    expire_date = datetime.fromtimestamp(
        int(extract_parameter_from_url("expire", base_url)), timezone.utc
    )
    expire_date_string = expire_date.astimezone().isoformat()
    result = f"This file is created with ytpb, and expires at {expire_date_string}"
    return result


def _compose_mpd_skeleton(playback, streams):
    nsmap = {
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        None: "urn:mpeg:DASH:schema:MPD:2011",
    }
    # Create the MPD tag
    mpd_element = etree.Element("MPD", nsmap=nsmap)

    mpd_element.attrib.update(
        {
            f"{{{nsmap['xsi']}}}schemaLocation": (
                "urn:mpeg:DASH:schema:MPD:2011 DASH-MPD.xsd"
            ),
            "profiles": "urn:mpeg:dash:profile:isoff-main:2011",
        }
    )

    some_base_url = next(iter(streams)).base_url

    comment_element = etree.Comment(_build_top_level_comment_text(some_base_url))
    mpd_element.addprevious(comment_element)

    # Create and append the ProgramInformation tag
    title, url = playback.info.title, playback.info.url
    program_info_element = E.ProgramInformation(E.Title(title), E.Source(url))
    mpd_element.append(program_info_element)

    # Create the Period tag
    period_element = etree.Element("Period")

    mime_types_of_streams = sorted({s.mime_type for s in streams})
    for i, mime_type in enumerate(mime_types_of_streams):
        adaptation_set_element = etree.Element("AdaptationSet")
        adaptation_set_element.attrib.update(
            {"id": str(i), "mimeType": mime_type, "subsegmentAlignment": "true"}
        )

        filtered_by_mime_type = streams.filter(lambda s: s.mime_type == mime_type)
        for stream in sorted(filtered_by_mime_type, key=lambda s: s.itag):
            representation_element = etree.Element("Representation")
            representation_element.attrib.update(
                {"id": stream.itag, "codecs": stream.codecs, "startWithSAP": "1"}
            )
            if "audio" in mime_type:
                representation_element.attrib.update(
                    {
                        "audioSamplingRate": str(stream.audio_sampling_rate),
                    }
                )
            else:
                representation_element.attrib.update(
                    {
                        "width": str(stream.width),
                        "height": str(stream.height),
                        "maxPlayoutRate": "1",
                        "frameRate": str(stream.frame_rate),
                    }
                )

            # Append the BaseURL tag
            representation_element.append(E("BaseURL", stream.base_url))

            # Append the Representation tag into AdaptationSet
            adaptation_set_element.append(representation_element)

        period_element.append(adaptation_set_element)

        # Finally, append the Period tag into MPD
        mpd_element.append(period_element)

    return mpd_element


def compose_static_mpd(
    playback: Playback, rewind_interval: RewindInterval, streams: SetOfStreams
) -> str:
    mpd_element = _compose_mpd_skeleton(playback, streams)

    some_base_url = next(iter(streams)).base_url

    rewind_length = rewind_interval.end.sequence - rewind_interval.start.sequence + 1
    segment_duration_s = float(extract_parameter_from_url("dur", some_base_url))
    range_duration_s = rewind_length * int(segment_duration_s)

    mpd_element.attrib.update(
        {"type": "static", "mediaPresentationDuration": f"PT{range_duration_s}S"}
    )

    period_element = mpd_element.find(".//Period", namespaces=NS)
    period_element.attrib["duration"] = f"PT{range_duration_s}S"

    segment_template_element = etree.Element("SegmentTemplate")
    segment_template_element.attrib.update(
        {
            "media": "sq/$Number$",
            "startNumber": str(rewind_interval.start.sequence),
            "duration": str(int(segment_duration_s) * int(S_TO_MS)),
            "timescale": "1000",
        }
    )

    for element in period_element.findall(".//AdaptationSet", namespaces=NS):
        element.insert(0, copy.deepcopy(segment_template_element))

    output = etree.tostring(
        etree.ElementTree(mpd_element),
        xml_declaration=True,
        encoding="UTF-8",
        pretty_print=True,
    )

    return output.decode()


def compose_dynamic_mpd(
    playback: Playback, rewind_metadata: SegmentMetadata, streams: SetOfStreams
) -> str:
    mpd_element = _compose_mpd_skeleton(playback, streams)
    mpd_element.attrib["profiles"] = "urn:mpeg:dash:profile:isoff-live:2011"

    mpd_element.attrib.update(
        {
            "type": "dynamic",
            "availabilityStartTime": datetime.fromtimestamp(
                rewind_metadata.ingestion_walltime, timezone.utc
            ).isoformat(timespec="milliseconds"),
        }
    )

    period_element = mpd_element.find(".//Period", namespaces=NS)
    period_element.attrib["start"] = f"PT{rewind_metadata.stream_duration}S"

    segment_template_element = etree.Element("SegmentTemplate")
    segment_template_element.attrib.update(
        {
            "media": "sq/$Number$",
            "startNumber": str(rewind_metadata.sequence_number),
            "timescale": "1000",
        }
    )

    segment_duration_ms = int(rewind_metadata.target_duration * S_TO_MS)
    segment_template_element.append(
        E("SegmentTimeline", E("S", d=str(segment_duration_ms), r="-1"))
    )

    for element in period_element.findall(".//AdaptationSet", namespaces=NS):
        element.insert(0, copy.deepcopy(segment_template_element))

    output = etree.tostring(
        etree.ElementTree(mpd_element),
        xml_declaration=True,
        encoding="UTF-8",
        pretty_print=True,
    )

    return output.decode()


def refresh_mpd(manifest_content: str, streams: SetOfStreams) -> str:
    manifest = etree.fromstring(manifest_content.encode())

    representation_elements = manifest.xpath("//mpd:Representation", namespaces=NS)
    for representation in representation_elements:
        base_url_element = representation.xpath(".//mpd:BaseURL", namespaces=NS)[0]
        if stream := streams.get_by_itag(itag := representation.get("id")):
            base_url_element.text = stream.base_url
        else:
            raise YtpbError(f"No stream with itag '{itag}' in the streams")

    comment_element: etree._Comment = manifest.xpath("/comment()")[0]
    comment_element.text = _build_top_level_comment_text(base_url_element.text)
    manifest.addprevious(comment_element)

    output = etree.tostring(
        etree.ElementTree(manifest),
        xml_declaration=True,
        encoding="UTF-8",
        pretty_print=True,
    )
    return output.decode()



import email
import shutil
import sys
from datetime import datetime, timedelta
from pathlib import Path
from typing import Literal

import click
import cloup
import structlog
from PIL import Image
from rich.console import Console
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)
from timedelta_isoformat import timedelta as isotimedelta

from ytpb.actions.capture import capture_frames, extract_frame_as_image
from ytpb.cli.commands.download import render_download_output_path_context
from ytpb.cli.common import (
    create_playback,
    get_parameter_by_name,
    prepare_line_for_summary_info,
    raise_for_sequence_ahead_of_current,
    raise_for_too_far_sequence,
    resolve_output_path,
    stream_argument,
)
from ytpb.cli.options import (
    cache_options,
    interval_option,
    no_cleanup_option,
    preview_option,
    validate_image_output_path,
    yt_dlp_option,
)
from ytpb.cli.parameters import (
    FormatSpecParamType,
    FormatSpecType,
    InputRewindInterval,
    PointInStreamParamType,
)
from ytpb.exceptions import QueryError, SegmentDownloadError, SequenceLocatingError
from ytpb.locate import SegmentLocator
from ytpb.segment import Segment
from ytpb.types import (
    AbsolutePointInStream,
    AddressableMappingProtocol,
    DateInterval,
    SegmentSequence,
)
from ytpb.utils.date import DurationFormatPattern, format_duration, ISO8601DateFormatter
from ytpb.utils.other import resolve_relativity_in_interval, S_TO_MS
from ytpb.utils.path import (
    expand_template_output_path,
    IntervalOutputPathContext,
    MinimalOutputPathContext,
    OUTPUT_PATH_PLACEHOLDER_RE,
    OutputPathContextRenderer,
    render_minimal_output_path_context,
)
from ytpb.utils.remote import request_reference_sequence

logger = structlog.get_logger(__name__)


class CaptureOutputPathContext(MinimalOutputPathContext):
    moment_date: datetime


class TimelapseOutputPathContext(MinimalOutputPathContext, IntervalOutputPathContext):
    every: timedelta


def render_capture_output_path_context(
    context: CaptureOutputPathContext,
    config_settings: AddressableMappingProtocol,
) -> CaptureOutputPathContext:
    output = context
    for variable in CaptureOutputPathContext.__annotations__.keys():
        match variable:
            case "moment_date" as x:
                date_formatter = ISO8601DateFormatter()
                output[x] = OutputPathContextRenderer.render_date(
                    context[x], date_formatter, config_settings
                )

    output.update(render_minimal_output_path_context(context, config_settings))

    return output


def render_timelapse_output_path_context(
    context: TimelapseOutputPathContext,
    config_settings: AddressableMappingProtocol,
) -> TimelapseOutputPathContext:
    output = context
    for variable in TimelapseOutputPathContext.__annotations__.keys():
        match variable:
            case "every" as x:
                output[x] = isotimedelta.isoformat(context[x]).replace("P", "E")

    output.update(render_download_output_path_context(context, config_settings))

    return output


def convert_every_option_to_timedelta(
    ctx: click.Context, param: click.Option, value: str
) -> timedelta:
    try:
        output = isotimedelta.fromisoformat(f"PT{value}")
    except ValueError:
        raise click.BadParameter(
            "Couldn't match repetetion format, e.g.: '30S', '1DT2H3M4S', etc."
        )
    return output


def print_timelapse_summary_info(
    dates: list[datetime], end_date: datetime, every: timedelta
) -> None:
    click.echo(
        "Every {} a time-lapse frame will be captured:".format(
            format_duration(every, DurationFormatPattern.IN_SENTENCE)
        )
    )

    def _format_datetime(x):
        return email.utils.format_datetime(x).split(", ")[1]

    click.echo("{:>12}: {} / S".format("Frame 1", _format_datetime(dates[0])))

    number_of_dates = len(dates)
    if number_of_dates == 2:
        click.echo("{:>12}: {}".format("Last frame", _format_datetime(dates[-1])))
    elif number_of_dates > 2:
        click.echo("{:>12}: {}".format("Frame 2", _format_datetime(dates[1])))
        if number_of_dates > 3:
            click.echo(f"{'':12}  {'...':^26}")
        click.echo("{:>12}: {}".format("Last frame", _format_datetime(dates[-1])))

    click.echo("{:>12}  {} / E".format("", _format_datetime(end_date)))


@cloup.group("capture", help="Capture a single or many frames.")
def capture_group():
    pass


@capture_group.command("frame", help="Capture a single frame.")
@cloup.option_group(
    "Input options",
    cloup.option(
        "-m",
        "--moment",
        metavar="MOMENT",
        type=PointInStreamParamType(allowed_literals=["now"]),
        help="Moment to capture.",
    ),
    cloup.option(
        "-vf",
        "--video-format",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.VIDEO),
        help="Video format to capture.",
    ),
)
@click.option(
    "-o",
    "--output",
    "output_path",
    type=click.Path(path_type=Path),
    help="Output path (with extension).",
    callback=validate_image_output_path(CaptureOutputPathContext),
)
@yt_dlp_option
@no_cleanup_option
@cache_options
@stream_argument
@click.pass_context
def frame_command(
    ctx: click.Context,
    moment: AbsolutePointInStream | Literal["now"],
    video_format: str,
    output_path: Path,
    yt_dlp: bool,
    no_cleanup: bool,
    force_update_cache: bool,
    no_cache: bool,
    stream_url: str,
):
    playback = create_playback(ctx)

    if video_format:
        logger.debug("Query video stream by format spec", spec=video_format)
        try:
            queried_video_streams = playback.streams.query(video_format)
        except QueryError as exc:
            click.echo(f"\nerror: Failed to query video streams. {exc}", err=True)
            sys.exit(1)
    else:
        queried_video_streams = []

    click.echo()
    click.echo("(<<) Locating a moment in the stream... ", nl=False)

    reference_stream = queried_video_streams[0]
    reference_base_url = reference_stream.base_url

    head_sequence = request_reference_sequence(reference_base_url, playback.session)

    match moment:
        case SegmentSequence() as sequence:
            raise_for_too_far_sequence(
                sequence, head_sequence, reference_base_url, ctx, "moment"
            )
            raise_for_sequence_ahead_of_current(sequence, head_sequence, ctx, "moment")
            moment_sequence = sequence
        case datetime() as date:
            try:
                sl = SegmentLocator(
                    reference_base_url,
                    temp_directory=playback.get_temp_directory(),
                    session=playback.session,
                )
                moment_sequence, _ = sl.find_sequence_by_time(date.timestamp())
            except SequenceLocatingError:
                message = "\nerror: An error occured during segment locating, exit."
                click.echo(message, err=True)
                sys.exit(1)
        case "now":
            moment_sequence = head_sequence

    click.echo("done.")

    try:
        segment_path = playback.download_segment(moment_sequence, reference_base_url)
        moment_segment = Segment.from_file(segment_path)
    except SegmentDownloadError as exc:
        click.echo()
        logger.error(exc, sequence=exc.sequence, exc_info=True)
        sys.exit(1)

    requested_moment_date: datetime
    match moment:
        case SegmentSequence() | "now":
            requested_moment_date = moment_segment.ingestion_start_date
        case datetime() as date:
            requested_moment_date = date

    # TODO: This should be expanded to take into account a case when a requested
    # date fall into a gap.
    actual_moment_date = requested_moment_date

    actual_moment_info_line = prepare_line_for_summary_info(
        actual_moment_date, actual_moment_date - requested_moment_date
    )

    click.echo(
        "Actual moment: {}, seq. {}".format(
            actual_moment_info_line, moment_segment.sequence
        )
    )

    # Absolute output path of an image with extension.
    final_output_path: Path
    if OUTPUT_PATH_PLACEHOLDER_RE.search(str(output_path)):
        template_context: CaptureOutputPathContext = {
            "id": playback.video_id,
            "title": playback.info.title,
            "moment_date": requested_moment_date,
        }
        final_output_path = expand_template_output_path(
            output_path,
            template_context,
            render_capture_output_path_context,
            ctx.obj.config,
        )
    else:
        final_output_path = output_path
    final_output_path = resolve_output_path(final_output_path)

    image = extract_frame_as_image(moment_segment, requested_moment_date)
    image.save(final_output_path, quality=80)

    try:
        saved_to_path_value = final_output_path.relative_to(Path.cwd())
    except ValueError:
        saved_to_path_value = final_output_path
    click.echo(f"\nSuccess! Saved to '{saved_to_path_value}'.")

    run_temp_directory = playback.get_temp_directory()
    if no_cleanup:
        click.echo(f"notice: No cleanup enabled, check {run_temp_directory}/")
    else:
        try:
            shutil.rmtree(run_temp_directory)
        except OSError:
            logger.warning(
                "Failed to remove %s temporary directory", run_temp_directory
            )


@capture_group.command("timelapse", help="Capture time-lapse frames.")
@cloup.option_group(
    "Input options",
    interval_option,
    cloup.option(
        "-e",
        "--every",
        metavar="REPETITION",
        help="Capture frames every REPETITION.",
        type=str,
        callback=convert_every_option_to_timedelta,
        required=True,
    ),
    cloup.option(
        "-vf",
        "--video-format",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.VIDEO),
        help="Video format to capture.",
    ),
    preview_option,
)
@click.option(
    "-o",
    "--output",
    "output_path",
    type=click.Path(path_type=Path),
    help="Output path (with extension).",
    callback=validate_image_output_path(TimelapseOutputPathContext),
)
@yt_dlp_option
@no_cleanup_option
@cache_options
@stream_argument
@click.pass_context
def timelapse_command(
    ctx: click.Context,
    interval: InputRewindInterval,
    every: timedelta,
    video_format: str,
    preview: bool,
    output_path: Path,
    yt_dlp: bool,
    no_cleanup: bool,
    force_update_cache: bool,
    no_cache: bool,
    stream_url: str,
):
    playback = create_playback(ctx)

    if video_format:
        logger.debug("Query video stream by format spec", spec=video_format)
        try:
            queried_video_streams = playback.streams.query(video_format)
        except QueryError as exc:
            click.echo(f"\nerror: Failed to query video streams. {exc}", err=True)
            sys.exit(1)
    else:
        queried_video_streams = []

    click.echo()
    click.echo("(<<) Locating interval start (S) and end (E)... ", nl=False)

    reference_stream = queried_video_streams[0]
    reference_base_url = reference_stream.base_url

    head_sequence = request_reference_sequence(reference_base_url, playback.session)

    requested_start, requested_end = resolve_relativity_in_interval(*interval)

    if isinstance(requested_start, SegmentSequence):
        raise_for_too_far_sequence(
            requested_start, head_sequence, reference_base_url, ctx, "interval"
        )
        raise_for_sequence_ahead_of_current(
            requested_start, head_sequence, ctx, "interval"
        )

    match requested_end:
        case SegmentSequence() as x:
            raise_for_sequence_ahead_of_current(x, head_sequence, ctx, "interval")
        case "now":
            requested_end = head_sequence - 1
        case "..":
            if not preview:
                raise click.BadParameter(
                    "Open interval is only valid in the preview mode",
                    param=get_parameter_by_name("interval", ctx),
                )

    try:
        rewind_interval = playback.locate_interval(
            requested_start, requested_end, reference_stream.itag
        )
    except SequenceLocatingError:
        message = "\nerror: An error occured during segment locating, exit."
        click.echo(message, err=True)
        sys.exit(1)

    click.echo("done.")

    start_segment = playback.get_downloaded_segment(
        rewind_interval.start.sequence, reference_base_url
    )
    end_segment = playback.get_downloaded_segment(
        rewind_interval.end.sequence, reference_base_url
    )

    requested_start_date: datetime
    match requested_start:
        case SegmentSequence():
            requested_start_date = start_segment.ingestion_start_date
        case datetime():
            requested_start_date = requested_start
        case timedelta() as delta:
            requested_start_date = end_segment.ingestion_end_date - delta

    requested_end_date: datetime
    match requested_end:
        case SegmentSequence():
            requested_end_date = end_segment.ingestion_end_date
        case datetime():
            requested_end_date = requested_end
        case timedelta() as delta:
            requested_end_date = start_segment.ingestion_start_date + delta

    requested_date_interval = DateInterval(requested_start_date, requested_end_date)
    actual_date_interval = DateInterval(
        start_segment.ingestion_start_date,
        end_segment.ingestion_end_date,
    )

    start_diff, end_diff = requested_date_interval - actual_date_interval

    cut_at_start_s = start_diff if start_diff > 0 else 0
    cut_at_end_s = abs(end_diff) if end_diff < 0 else 0
    {
        "cut_at_start": round(cut_at_start_s * int(S_TO_MS)),
        "cut_at_end": round(cut_at_end_s * int(S_TO_MS)),
    }
    actual_date_interval = DateInterval(
        start=actual_date_interval.start + timedelta(seconds=cut_at_start_s),
        end=actual_date_interval.end - timedelta(seconds=cut_at_end_s),
    )

    s = requested_date_interval.start
    e = requested_date_interval.end
    dates_to_capture = [s + every * i for i in range((e - s) // every + 1)]

    print_timelapse_summary_info(dates_to_capture, requested_date_interval.end, every)

    if preview:
        click.echo("info: Preview mode is enabled, only first 3 frames will be taken.")
        dates_to_capture = dates_to_capture[:3]

    click.echo()

    # Absolute output path of images with a numeric pattern.
    final_output_path: Path
    if OUTPUT_PATH_PLACEHOLDER_RE.search(str(output_path)):
        input_timezone = requested_date_interval.start.tzinfo
        template_context: TimelapseOutputPathContext = {
            "id": playback.video_id,
            "title": playback.info.title,
            "input_start_date": requested_date_interval.start,
            "input_end_date": requested_date_interval.end,
            "actual_start_date": actual_date_interval.start.astimezone(input_timezone),
            "actual_end_date": actual_date_interval.end.astimezone(input_timezone),
            "duration": requested_end_date - requested_start_date,
            "every": every,
        }
        final_output_path = expand_template_output_path(
            output_path,
            template_context,
            render_timelapse_output_path_context,
            ctx.obj.config,
        )
    else:
        final_output_path = output_path
    final_output_path = resolve_output_path(final_output_path)

    click.echo("(<<) Capturing frames as images:")

    capturing_progress = Progress(
        TextColumn("{task.description}"),
        BarColumn(bar_width=32),
        MofNCompleteColumn(),
        TaskProgressColumn(),
        TextColumn("eta"),
        TimeRemainingColumn(),
        console=Console(),
    )
    capturing_task = capturing_progress.add_task("", total=len(dates_to_capture))

    def _save_ith_frame_as_image(
        image: Image, output_path_pattern: Path, i: int
    ) -> None:
        image_output_name = output_path_pattern.name % i
        image_output_path = output_path_pattern.with_name(image_output_name)
        image.save(image_output_path, quality=80)

    with capturing_progress:
        # Save the first frame of a time-lapse -- the start segment was already located.
        first_frame_image = extract_frame_as_image(start_segment, requested_start_date)
        _save_ith_frame_as_image(first_frame_image, final_output_path, 0)
        capturing_progress.advance(capturing_task)

        captured = capture_frames(
            playback, dates_to_capture[1:], reference_base_url, start_segment.sequence
        )
        for i, (image, _) in enumerate(captured, 1):
            _save_ith_frame_as_image(image, final_output_path, i)
            capturing_progress.advance(capturing_task)

    try:
        saved_to_path_value = final_output_path.parent.relative_to(Path.cwd())
    except ValueError:
        saved_to_path_value = final_output_path.parent
    click.echo(f"\nSuccess! Saved to '{saved_to_path_value}/'.")

    run_temp_directory = playback.get_temp_directory()
    if no_cleanup:
        click.echo(f"notice: No cleanup enabled, check {run_temp_directory}/")
    else:
        try:
            shutil.rmtree(run_temp_directory)
        except OSError:
            logger.warning(
                "Failed to remove %s temporary directory", run_temp_directory
            )



"""TODO: The current implementation of segment merging is not optimal in terms
of disk space, as it requires triple the amount of the total size of
segments. Should be changed to use different techniques, without the need for
intermediate files.
"""

import os
import shlex
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any

from ytpb import ffmpeg
from ytpb.utils.other import S_TO_MS

DEFAULT_VIDEO_ENCODING_SETTINGS = {
    "h264": "libx264 -crf 18",
    "vp9": "libvpx-vp9 -crf 31 -b:v 0",
    "av1": "libaom-av1 -crf 31",
}


def get_nth_or_none(iterable, n: int) -> Any | None:
    if not iterable:
        return None
    return iterable[n]


def mux_and_cut_boundary_segment(
    audio_segment_path: Path, video_segment_path: Path, output_path: Path, **cut_kwargs
):
    def prepare_ffmpeg_input_options(
        segment_path: Path, cut_at_start: int = 0, cut_at_end: int = 0
    ):
        assert cut_at_start or cut_at_end
        if cut_at_start > 0:
            return ["-ss", f"{cut_at_start}ms", "-i", segment_path]
        elif cut_at_end > 0:
            with NamedTemporaryFile(suffix=segment_path.suffix) as f:
                temp_path = Path(f.name)
                ffmpeg.ffmpeg_stream_copy(segment_path, temp_path)
                segment_duration_s = float(
                    ffmpeg.ffprobe_show_entries(temp_path, "format=duration")
                )
            end_seek_pos: int = round(segment_duration_s * S_TO_MS) - cut_at_end
            return ["-i", segment_path, "-to", f"{end_seek_pos}ms"]
        else:
            return ["-i", segment_path]

    if not {"cut_at_start", "cut_at_end"} > cut_kwargs.keys():
        raise ValueError(
            "only cut_at_start or cut_at_end keyword argument is accepted in cut_kwargs"
        )

    ffmpeg_input_options = []
    ffmpeg_codecs_options = []

    should_cut = bool(
        cut_kwargs.get("cut_at_start", 0) or cut_kwargs.get("cut_at_end", 0)
    )
    if should_cut is False:
        if video_segment_path:
            ffmpeg_input_options += ["-i", str(video_segment_path)]
        if audio_segment_path:
            ffmpeg_input_options += ["-i", str(audio_segment_path)]
        ffmpeg.run_ffmpeg(ffmpeg_input_options + ["-c", "copy", str(output_path)])
    else:
        if video_segment_path:
            ffmpeg_input_options += prepare_ffmpeg_input_options(
                video_segment_path, **cut_kwargs
            )

            video_codec_name = ffmpeg.ffprobe_show_entries(
                video_segment_path, "stream=codec_name"
            )

            user_encoding_settings = os.environ.get(
                f"YTPB_{video_codec_name.upper()}_ENCODING_SETTINGS", None
            )
            if user_encoding_settings:
                ffmpeg_codecs_options += ["-c:v"] + shlex.split(user_encoding_settings)
            else:
                try:
                    ffmpeg_codecs_options += ["-c:v"] + shlex.split(
                        DEFAULT_VIDEO_ENCODING_SETTINGS[video_codec_name]
                    )
                except KeyError:
                    raise ValueError(
                        "No encoding settings are availabe for"
                        f"'{video_codec_name}' video codec"
                    )

        if audio_segment_path:
            ffmpeg_input_options += prepare_ffmpeg_input_options(
                audio_segment_path, **cut_kwargs
            )
            ffmpeg_codecs_options += ["-c:a", "copy"]

        ffmpeg.run_ffmpeg(
            ffmpeg_input_options + ffmpeg_codecs_options + [str(output_path)]
        )


def _compose_concat_file(segment_paths, temp_directory, suffix: str = ""):
    if suffix:
        concat_file_path = Path(temp_directory, f"concat_{suffix}")
    else:
        concat_file_path = Path(temp_directory, "concat")

    concat_file_lines = [f"file '{path}'\n" for path in segment_paths]

    with open(concat_file_path, "w") as f:
        f.writelines(concat_file_lines)

    return concat_file_path


def ensure_cleanup_if_needed(f):
    def g(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        finally:
            if kwargs.get("cleanup", True):
                assert getattr(g, "paths_to_cleanup", None) is not None
                for path in g.paths_to_cleanup:
                    path.unlink(missing_ok=True)

    return g


@ensure_cleanup_if_needed
def merge_segments(
    audio_segment_paths: list[Path] | None = None,
    video_segment_paths: list[Path] | None = None,
    output_directory: str | Path | None = None,
    output_stem: str | Path | None = None,
    temp_directory: str | Path | None = None,
    cut_at_start: int = 0,
    cut_at_end: int = 0,
    cleanup: bool = True,
) -> Path:
    merge_segments.paths_to_cleanup = []

    if audio_segment_paths is None and video_segment_paths is None:
        raise ValueError("Audio or/and video paths should be provided")

    audio_segment_paths = audio_segment_paths or []
    video_segment_paths = video_segment_paths or []

    if audio_segment_paths and video_segment_paths:
        if len(audio_segment_paths) != len(video_segment_paths):
            raise ValueError("Audio and video paths are not of equal size")
        output_extension = ".mkv"
    elif audio_segment_paths:
        output_extension = os.path.splitext(audio_segment_paths[0])[1]
    elif video_segment_paths:
        output_extension = os.path.splitext(video_segment_paths[0])[1]

    if output_directory:
        output_directory = Path(output_directory)
    else:
        output_directory = Path.cwd()
    output_path = output_directory / f"{output_stem}{output_extension}"

    safe_concat_options = ["-safe", "0", "-f", "concat"]

    should_cut = bool(cut_at_start or cut_at_end)
    if should_cut is False:
        no_cut_concat_options = []
        if video_segment_paths:
            concat_file_path = _compose_concat_file(
                video_segment_paths, temp_directory, suffix="video"
            )
            merge_segments.paths_to_cleanup.append(concat_file_path)
            no_cut_concat_options += safe_concat_options + ["-i", concat_file_path]
        if audio_segment_paths:
            concat_file_path = _compose_concat_file(
                audio_segment_paths, temp_directory, suffix="audio"
            )
            merge_segments.paths_to_cleanup.append(concat_file_path)
            no_cut_concat_options += safe_concat_options + ["-i", concat_file_path]
        ffmpeg.run_ffmpeg(no_cut_concat_options + ["-c", "copy", output_path])
    else:
        num_of_segments = len(audio_segment_paths or video_segment_paths)

        start_audio_segment_path = get_nth_or_none(audio_segment_paths, 0)
        start_video_segment_path = get_nth_or_none(video_segment_paths, 0)

        end_audio_segment_path = get_nth_or_none(audio_segment_paths, -1)
        end_video_segment_path = get_nth_or_none(video_segment_paths, -1)

        if num_of_segments == 1:
            segment_trimmed_path = Path(temp_directory, "a.a" + output_extension)
            mux_and_cut_boundary_segment(
                start_audio_segment_path,
                start_video_segment_path,
                segment_trimmed_path,
                cut_at_start=cut_at_start,
            )
            parts_to_merge = [segment_trimmed_path]
        elif num_of_segments == 2:
            start_trimmed_path = Path(temp_directory, "ab.a" + output_extension)
            mux_and_cut_boundary_segment(
                start_audio_segment_path,
                start_video_segment_path,
                start_trimmed_path,
                cut_at_start=cut_at_start,
            )

            end_trimmed_path = start_trimmed_path.with_stem("ab.b")
            mux_and_cut_boundary_segment(
                end_audio_segment_path,
                end_video_segment_path,
                end_trimmed_path,
                cut_at_end=cut_at_end,
            )

            parts_to_merge = [start_trimmed_path, end_trimmed_path]
        else:
            start_trimmed_path = Path(temp_directory, "abc.a" + output_extension)
            mux_and_cut_boundary_segment(
                start_audio_segment_path,
                start_video_segment_path,
                start_trimmed_path,
                cut_at_start=cut_at_start,
            )

            middle_concatenated_path = start_trimmed_path.with_stem("abc.b")

            end_trimmed_path = start_trimmed_path.with_stem("abc.c")
            mux_and_cut_boundary_segment(
                end_audio_segment_path,
                end_video_segment_path,
                end_trimmed_path,
                cut_at_end=cut_at_end,
            )

            bounds_slice = slice(1, -1)
            concat_filter_options = []

            if video_segment_paths:
                concat_file_path = _compose_concat_file(
                    video_segment_paths[bounds_slice],
                    temp_directory,
                    suffix="video",
                )
                concat_filter_options += safe_concat_options + ["-i", concat_file_path]
                merge_segments.paths_to_cleanup.append(concat_file_path)

            if audio_segment_paths:
                concat_file_path = _compose_concat_file(
                    audio_segment_paths[bounds_slice],
                    temp_directory,
                    suffix="audio",
                )
                concat_filter_options += safe_concat_options + ["-i", concat_file_path]
                merge_segments.paths_to_cleanup.append(concat_file_path)

            ffmpeg.run_ffmpeg(
                concat_filter_options + ["-c", "copy", middle_concatenated_path],
                capture_output=True,
                check=True,
            )

            parts_to_merge = [
                start_trimmed_path,
                middle_concatenated_path,
                end_trimmed_path,
            ]

        merge_segments.paths_to_cleanup.extend(parts_to_merge)

        if num_of_segments == 1:
            ffmpeg.run_ffmpeg(
                ["-i", str(parts_to_merge[0]), "-c", "copy", str(output_path)]
            )
        else:
            concat_file_path = _compose_concat_file(parts_to_merge, temp_directory)
            ffmpeg.run_ffmpeg(
                safe_concat_options
                + ["-i", str(concat_file_path), "-c", "copy", str(output_path)]
            )
            merge_segments.paths_to_cleanup.append(concat_file_path)

    return output_path



import io
import os
from pathlib import Path
from typing import Callable
from urllib.parse import urljoin

import requests
import structlog

from ytpb.exceptions import SegmentDownloadError
from ytpb.types import SegmentSequence
from ytpb.utils.url import extract_media_type_from_url, extract_parameter_from_url

logger = structlog.get_logger(__name__)

_SegmentOutputFilename = str | Callable[[SegmentSequence, str], str]


def compose_default_segment_filename(sequence: SegmentSequence, base_url: str) -> str:
    itag = extract_parameter_from_url("itag", base_url)
    extension = extract_media_type_from_url(base_url)[1]
    return f"{sequence}.i{itag}.{extension}"


def _make_request_for_segment(
    sequence: SegmentSequence,
    base_url: str,
    size: int | None = None,
    session: requests.Session | None = None,
) -> requests.Response:
    get_function = session.get if session else requests.get

    headers = {}
    if size:
        headers["Range"] = f"bytes=0-{size}"

    response = get_function(urljoin(base_url, f"sq/{sequence}"), headers=headers)

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError:
        exception = SegmentDownloadError(
            f"Failed to download segment {sequence}: {response.reason}", sequence
        )
        logger.error(exception, exc_info=True)
        raise exception from None

    if size and response.status_code == requests.codes.ok:
        logger.debug("Header 'Range' is ignored, downloaded all content")

    return response


def download_segment(
    sequence: int,
    base_url: str,
    output_directory: Path,
    output_filename: _SegmentOutputFilename = compose_default_segment_filename,
    size: int | None = None,
    session: requests.Session | None = None,
    force_download: bool = True,
) -> Path:
    if callable(output_filename):
        output_filename_value = output_filename(sequence, base_url)
        path_to_download_to = Path(output_directory) / output_filename_value
    else:
        path_to_download_to = Path(output_directory) / output_filename

    if force_download or not os.path.isfile(path_to_download_to):
        with open(path_to_download_to, "wb") as f:
            response = _make_request_for_segment(sequence, base_url, size, session)
            f.write(response.content)

    return path_to_download_to


def download_segment_to_buffer(
    sequence: SegmentSequence,
    base_url: str,
    size: int | None = None,
    session: requests.Session | None = None,
) -> io.BytesIO:
    response = _make_request_for_segment(sequence, base_url, size, session)
    return io.BytesIO(response.content)



import logging
import operator
import os
import re
import tomllib
from collections import ChainMap
from functools import reduce
from itertools import product
from pathlib import Path
from typing import Any, Callable

import structlog
from platformdirs import user_config_path


class AddressableMixin:
    def traverse(self, address: str, default: Any = None, delimiter: str = ".") -> Any:
        """Traverse nested dictionary to access a value by an address."""
        try:
            keys = address.split(delimiter)
            return reduce(operator.getitem, keys, self)
        except KeyError:
            return default


class AddressableDict(dict, AddressableMixin):
    """A dictionary that allows to use an address to access a nested value."""


class AddressableChainMap(ChainMap, AddressableMixin):
    """ChainMap that allows to use an address to access a nested value."""


USER_AGENT = "Mozilla/5.0 (Android 14; Mobile; rv:68.0) Gecko/68.0 Firefox/120.0"


# Dynamic aliases (such aliases that are expanded by functions).
def expand_itag_aliases(format_spec: str) -> str:
    """@<itag> = itag eq <itag>"""
    return re.sub(r"@(\d+)\b", r"itag eq \1", format_spec)


ALIAS_EXPAND_FUNCTIONS: tuple[Callable[[str], str]] = (expand_itag_aliases,)

# Static aliases (such aliases that are explicitly defined).
FORMAT_ALIASES = {"mp4": "format eq mp4", "webm": "format eq webm"}

VIDEO_QUALITY_HEIGHTS = (144, 240, 360, 480, 720, 1080, 1440, 2160)

VIDEO_QUALITY_30FPS_ALIASES = {
    f"{height}p": f"height eq {height} and frame_rate eq 30"
    for height in VIDEO_QUALITY_HEIGHTS
}
VIDEO_QUALITY_30FPS_ALIASES.update(
    {
        f"{height}p30": f"height eq {height} and frame_rate eq 30"
        for height in VIDEO_QUALITY_HEIGHTS
    }
)

VIDEO_QUALITY_60FPS_ALIASES = {
    f"{height}p60": f"height eq {height} and frame_rate eq 60"
    for height in [720, 1080, 1440, 2160]
}

VIDEO_QUALITY_WITH_OPERATOR_ALIASES = {
    f"{operator}{height}p": f"height {operator_name} {height}"
    for (operator, operator_name), height in product(
        [("<", "lt"), ("<=", "le"), (">", "gt"), (">=", "ge")], VIDEO_QUALITY_HEIGHTS
    )
}

NAME_QUALITY_ALIASES = {
    "low": "height eq 144",
    "medium": "height eq 480",
    "high": "height eq 720",
    "FHD": "height eq 1080",
    "2K": "height eq 1440",
    "4K": "height eq 2160",
}

FRAME_PER_SECOND_ALIASES = {"30fps": "frame_rate eq 30", "60fps": "frame_rate eq 60"}

BUILT_IN_ALIASES = {
    **FORMAT_ALIASES,
    **VIDEO_QUALITY_30FPS_ALIASES,
    **VIDEO_QUALITY_60FPS_ALIASES,
    **VIDEO_QUALITY_WITH_OPERATOR_ALIASES,
    **NAME_QUALITY_ALIASES,
    **FRAME_PER_SECOND_ALIASES,
}

ALL_ALIASES = {**BUILT_IN_ALIASES}

DEFAULT_OUTPUT_PATH = "<title>_<input_start_date>"

DEFAULT_CONFIG = AddressableDict(
    {
        "version": 1,
        "options": {
            "download": {
                "audio_format": "itag eq 140",
                "video_format": (
                    "best(format eq mp4 and height le 1080 and frame_rate eq 30)"
                ),
                "output_path": DEFAULT_OUTPUT_PATH,
            },
            "capture": {
                "frame": {
                    "video_format": "best(format eq mp4 and frame_rate eq 30)",
                    "output_path": "<title>_<moment_date>.jpg",
                },
                "timelapse": {
                    "video_format": "best(format eq mp4 and frame_rate eq 30)",
                    "output_path": (
                        "<title>/<input_start_date>/<every>/"
                        "<title>_<input_start_date>_<every>_%04d.jpg"
                    ),
                },
            },
            "mpd": {
                "compose": {
                    "audio_formats": "itag eq 140",
                    "video_formats": (
                        "format eq webm and [height eq 720 or height eq 1080] and "
                        "frame_rate eq 30"
                    ),
                    "output_path": f"{DEFAULT_OUTPUT_PATH}.mpd",
                }
            },
        },
        "general": {
            "preview_duration": 10,
            "user_agent": USER_AGENT,
        },
        "output": {
            "date": {"styles": "basic,complete,hh"},
            "title": {
                "style": "custom",
                "custom": {
                    "characters": "posix",
                    "separator": "-",
                    "max_length": 30,
                    "lowercase": False,
                },
            },
        },
    }
)


def setup_logging(level: int) -> None:
    structlog.configure(
        processors=[
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.TimeStamper(fmt="%s.%f", utc=True),
            structlog.dev.ConsoleRenderer(
                exception_formatter=structlog.dev.plain_traceback,
                # For details about NO_COLOR, see https://no-color.org/
                colors="NO_COLOR" not in os.environ,
            ),
            structlog.dev.set_exc_info,
        ],
        wrapper_class=structlog.make_filtering_bound_logger(level),
        context_class=dict,
        logger_factory=structlog.PrintLoggerFactory(),
        cache_logger_on_first_use=True,
    )

    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("urllib3").propagate = False


def get_default_config_path() -> Path:
    return user_config_path("ytpb") / "config.toml"


def load_config_from_file(path: Path) -> dict:
    with path.open("rb") as f:
        return tomllib.load(f)


def update_nested_dict(base: dict, updates: dict) -> dict:
    """Update a base nested dict with values from updates. The new, update
    dictionary will be returned.
    """
    updated_dict = base.copy()
    for key, value in updates.items():
        if isinstance(value, dict) and isinstance(base.get(key), dict):
            updated_dict[key] = update_nested_dict(base[key], value)
        else:
            updated_dict[key] = value
    return updated_dict



from datetime import datetime, timedelta, timezone

import click
import pytest
from dateutil.tz import tzlocal
from freezegun import freeze_time
from freezegun.api import FakeDatetime

from ytpb.cli.parameters import (
    FormatSpecParamType,
    FormatSpecType,
    RewindIntervalParamType,
)


def test_none_format_spec():
    spec_type = FormatSpecType.AUDIO
    assert FormatSpecParamType(spec_type).convert("none", None, None) is None
    assert FormatSpecParamType(spec_type).convert("", None, None) is None


def test_format_spec_without_function():
    expected = "mime_type contains audio and [format eq mp4]"
    actual = FormatSpecParamType(FormatSpecType.AUDIO).convert(
        "format eq mp4", None, None
    )
    assert expected == actual


def test_format_spec_with_function():
    expected = "function(mime_type contains audio and [format eq mp4])"
    actual = FormatSpecParamType(FormatSpecType.AUDIO).convert(
        "function(format eq mp4)", None, None
    )
    assert expected == actual


@pytest.mark.parametrize(
    "value,expected",
    [
        ("0/100", (0, 100)),
        (
            "0/20240102T102030+00",
            (0, datetime(2024, 1, 2, 10, 20, 30, tzinfo=timezone.utc)),
        ),
        (
            "20240102T102000+00/100",
            (datetime(2024, 1, 2, 10, 20, tzinfo=timezone.utc), 100),
        ),
        ("0/P1DT30S", (0, timedelta(days=1, seconds=30))),
        ("P1DT30S/100", (timedelta(days=1, seconds=30), 100)),
        (
            "20240102T102000+00/20240102T102030+00",
            (
                datetime(2024, 1, 2, 10, 20, tzinfo=timezone.utc),
                datetime(2024, 1, 2, 10, 20, 30, tzinfo=timezone.utc),
            ),
        ),
        (
            "20240102T102000/20240102T102030",
            (
                datetime(2024, 1, 2, 10, 20),
                datetime(2024, 1, 2, 10, 20, 30),
            ),
        ),
        (
            "20240102T102000+00/P1DT30S",
            (
                datetime(2024, 1, 2, 10, 20, tzinfo=timezone.utc),
                timedelta(days=1, seconds=30),
            ),
        ),
        (
            "P1DT30S/20240102T102030+00",
            (
                timedelta(days=1, seconds=30),
                datetime(2024, 1, 2, 10, 20, 30, tzinfo=timezone.utc),
            ),
        ),
        (
            "20240102T102000+00/now",
            (datetime(2024, 1, 2, 10, 20, tzinfo=timezone.utc), "now"),
        ),
        ("PT30S/now", (timedelta(seconds=30), "now")),
        (
            "20240102T102000+00/..",
            (datetime(2024, 1, 2, 10, 20, tzinfo=timezone.utc), ".."),
        ),
    ],
)
def test_rewind_interval(value: str, expected):
    assert expected == RewindIntervalParamType().convert(value, None, None)


@pytest.mark.parametrize(
    "value,expected",
    [
        (
            "20240102T102000/T25M",
            (datetime(2024, 1, 2, 10, 20), (datetime(2024, 1, 2, 10, 25))),
        ),
        (
            "20240102T102000+00/T25M",
            (
                datetime(2024, 1, 2, 10, 20, tzinfo=timezone.utc),
                (datetime(2024, 1, 2, 10, 25, tzinfo=timezone.utc)),
            ),
        ),
        (
            "2023Y12M31DT25M/20240102T102030",
            (datetime(2023, 12, 31, 10, 25, 30), (datetime(2024, 1, 2, 10, 20, 30))),
        ),
    ],
)
def test_rewind_interval_with_replacing_components(value: str, expected):
    assert expected == RewindIntervalParamType().convert(value, None, None)


@pytest.mark.parametrize(
    "value,expected",
    [
        (
            "10:25/T1035",
            (
                FakeDatetime(2024, 1, 2, 10, 25, tzinfo=tzlocal()),
                FakeDatetime(2024, 1, 2, 10, 35, tzinfo=tzlocal()),
            ),
        ),
    ],
)
@freeze_time("2024-01-02T10:20:30-01")
def test_rewind_interval_with_time_of_today(value: str, expected):
    assert expected == RewindIntervalParamType().convert(value, None, None)


@pytest.mark.parametrize(
    "value,invalid_part",
    [
        ("20240102T102070+00/PT30M", "20240102T102070+00"),
        ("20240102T102030+00/PT30", "PT30"),
    ],
)
def test_invalid_rewind_interval(value: str, invalid_part: str):
    with pytest.raises(click.BadParameter) as exc_info:
        RewindIntervalParamType().convert(value, None, None)
    assert invalid_part in str(exc_info.value)



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Protocol, TypeAlias

from ytpb.config import AddressableChainMap
from ytpb.mpd import AudioRepresentationInfo, VideoRepresentationInfo


Timestamp: TypeAlias = float

SegmentSequence: TypeAlias = int


class RelativeSegmentSequence(int): ...


AbsolutePointInStream: TypeAlias = datetime | SegmentSequence
RelativePointInStream: TypeAlias = timedelta | RelativeSegmentSequence
PointInStream: TypeAlias = AbsolutePointInStream | RelativePointInStream

AudioStream: TypeAlias = AudioRepresentationInfo
VideoStream: TypeAlias = VideoRepresentationInfo
AudioOrVideoStream: TypeAlias = AudioStream | VideoStream
SetOfStreams: TypeAlias = "Streams[AudioOrVideoStream]"


class AddressableMappingProtocol(Protocol):
    def traverse(self, address: str, default: Any, delimiter: str) -> Any: ...


ConfigMap: TypeAlias = AddressableChainMap


@dataclass
class DateInterval:
    """Represents a closed date interval."""

    start: datetime
    end: datetime

    def __post_init__(self):
        if self.end < self.start:
            raise ValueError("End date occurs earlier in time than the start date")

    @property
    def duration(self) -> float:
        return (self.end - self.start).total_seconds()

    def __sub__(self, other: "DateInterval") -> tuple[float, float]:
        """Find difference between two date intervals as:
        [x1, y1] - [x2, y2] = [|x2 - x1|, |y2 - y1|].
        """
        return (
            (self.start - other.start).total_seconds(),
            (self.end - other.end).total_seconds(),
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DateInterval):
            return False
        return self.start == self.start and self.end == other.end

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __le__(self, other: "DateInterval") -> bool:
        return self.is_subinterval(other)

    def __contains__(self, value: datetime) -> bool:
        return self.start <= value <= self.end

    def __str__(self) -> str:
        return f"[{self.start.isoformat(), self.end.isoformat()}]"

    def is_subinterval(self, other: "DateInterval") -> bool:
        """Test whether the interval is completely included in other."""
        return self.start in other and self.end in other
